import numpy as np
from abc import ABC, abstractmethod
import matplotlib.patches as patches


class Region(ABC):
    def __init__(self, rng=None):
        if rng is not None:
            self.rng = rng
        else:
            self.rng = np.random.default_rng()
        
    @abstractmethod
    def is_in_region(self, coordinate):
        pass

    @abstractmethod
    def equally_spaced_points(self):
        pass

    @abstractmethod
    def sample_points(self, num_points):
        pass

    @abstractmethod
    def plot(self, ax, label=None):
        pass

class CombinedRegion(Region):
    def __init__(self, regions, rng=None):
        super().__init__(rng)
        self.regions = regions
        assert not self._overlaps(), "Not implemented for overlapping regions"

        self.volumes = np.array([reg.volume for reg in self.regions])
        self.volume = np.sum(self.volumes)

    def _overlaps(self):
        for r in self.regions:
            points_fixed = r.equally_spaced_points()
            points_sampled = r.sample_points(100)
            for other_reg in self.regions:
                if r is other_reg:
                    continue
                if np.any(other_reg.is_in_region(points_fixed)) or \
                    np.any(other_reg.is_in_region(points_sampled)):
                    return True
        return False

    def is_in_region(self, coordinate):
        return np.logical_or([reg.is_in_region(coordinate) for reg in self.regions])

    def equally_spaced_points(self):
        return np.concatenate([reg.equally_spaced_points() for reg in self.regions], axis=0)

    def sample_points(self, num_points):
        vol_ratio = self.volumes / self.volume
        sample_limit = np.cumsum(vol_ratio)
        test_val = self.rng.uniform(0,1,num_points)
        sample_result = test_val[None,:] < sample_limit[:,None]
        reg_idx = np.sum(sample_result, axis=0) - 1
        unique, counts = np.unique(reg_idx, return_counts=True)
        sampled_points = np.concatenate([self.regions[idx].sample_points(num) 
                        for idx, num in zip(unique, counts)])
        assert sampled_points.shape[0] == num_points
        return sampled_points

    def plot(self, ax, label=None):
        for reg in self.regions:
            reg.plot(ax, label)
            

class Cuboid(Region):
    def __init__(self, side_lengths, center=(0, 0, 0), point_spacing=(1,1,1), rng=None):
        super().__init__(rng)
        self.side_lengths = np.array(side_lengths)
        self.center = np.array(center)
        self.low_lim = self.center - self.side_lengths / 2
        self.high_lim = self.center + self.side_lengths / 2
        self.volume = np.prod(side_lengths)
        self.point_spacing = point_spacing

    def is_in_region(self, coordinates, padding=[[0,0,0]]):
        is_in_coord_wise = np.logical_and(coordinates >= self.low_lim[None,:]-padding,
                                        coordinates <= self.high_lim[None,:]+padding)
        is_in = np.logical_and(np.logical_and(is_in_coord_wise[:,0], 
                                                is_in_coord_wise[:,1]),
                                                is_in_coord_wise[:,2])
        return is_in

    def equally_spaced_points(self, point_dist=None):
        if point_dist is None:
            point_dist = self.point_spacing
        if isinstance(point_dist, (int, float)):
            point_dist = np.ones(3) * point_dist
        else:
            point_dist = np.array(point_dist)

        num_points = np.floor(self.side_lengths / point_dist)
        assert min(num_points) >= 1

        single_axes = [np.arange(num_p) * p_dist for num_p, p_dist in zip(num_points, point_dist)]
        remainders = [
            s_len - single_ax[-1]
            for single_ax, s_len in zip(single_axes, self.side_lengths)
        ]
        single_axes = [
            single_ax + remainder / 2
            for single_ax, remainder in zip(single_axes, remainders)
        ]
        all_points = np.meshgrid(*single_axes)
        all_points = np.concatenate(
            [points.flatten()[:, None] for points in all_points], axis=-1
        )
        all_points = all_points + self.low_lim
        return all_points

    def sample_points(self, num_points):
        low_lim = np.array(self.low_lim)
        high_lim = np.array(self.high_lim)

        samples = self.rng.uniform(low_lim[None,:], high_lim[None,:], size=(num_points, 3))

        # samples = [
        #     self.rng.uniform(low_lim, high_lim, (num_points, 1))
        #     for low_lim, high_lim in zip(self.low_lim, self.high_lim)
        # ]

        # samples = np.concatenate(samples, axis=-1)
        return samples

    def plot(self, ax, label=None):
        rect = patches.Rectangle(self.low_lim, self.side_lengths[0], self.side_lengths[1], fill=True, alpha=0.3, label=label)
        ax.add_patch(rect)


class Rectangle(Region):
    def __init__(self, side_lengths, center, point_spacing=(1,1), spatial_dim=3, rng=None):
        """Constructs a two-dimensional rectangle

        The rectangle can be placed in 3D space by setting a 3-dimensional center
        
        Parameters
        ----------
        side_lengths : array_like of shape (2,)
            Lengths of the sides of the rectangle.
        center : array_like of shape (2,) or (3,)
            Center of the cylinder. The length of the array determines the spatial dimension.
            If the array has length 3, the rectangle is placed in 3D space with the z-coordinate
            fixed at the value of the third element.
        point_spacing : array_like of shape (2,), optional
            Spacing between points in each direction, affects the selection of points
            for the equally_spaced_points method. The default is (1,1,1).
        spatial_dim : int, optional
            Spatial dimension of the rectangle. The default is 3. Must correspond to the length
            of the center array. 
        rng : np.random.Generator, optional
            Random number generator. The default is None, in which case a new generator
            with a random seed will be created. 
            The generator affects the sampling of points in the sample_points method, and so 
            should be supplied for a reproducible result.

        Returns
        -------
        rectangle : Rectangle

        Notes
        -----
        Although there are many more rectangles in 3D, here only rectangles with a fixed
        z-coordinate is considered. A rotation could in principle be applied to the rectangle
        for other orientations.
        """
        super().__init__(rng)
        assert spatial_dim in (2,3)
        assert len(center) == spatial_dim
        self.side_lengths =  np.array(side_lengths)
        self.center = np.array(center)
        self.point_spacing = np.array(point_spacing)
        self.spatial_dim = spatial_dim

        self.low_lim = self.center[:2] - self.side_lengths / 2
        self.high_lim = self.center[:2] + self.side_lengths / 2
        self.volume = np.prod(side_lengths)
        
    def is_in_region(self, coordinates, padding=[[0,0]]):
        """Checks whether the coordinate is within the region or not. 
        
        Parameters
        ----------
        coordinates : np.ndarray
            Shape (N,3) or (N, 2) where N is the number of coordinates to check. 
            Each row is a coordinate in 3D space.      
        padding : array_like of shape (2,), optional
            Padding to add to the region. The default is [[0,0]]. This can 
            be used to check whether a point is near the region, rather than strictly 
            inside it.
            
        Returns
        -------
        is_in : boolan np.ndarray of shape (N,)
            True if the coordinate is within the region, False otherwise.
        """
        if self.spatial_dim == 3:
            if not np.allclose(coordinates[:,2], self.center[2]):
                return False

        is_in_coord_wise = np.logical_and(coordinates[:,:2] >= self.low_lim[None,:]-padding,
                                        coordinates[:,:2] <= self.high_lim[None,:]+padding)
        is_in = np.logical_and(is_in_coord_wise[:,0], is_in_coord_wise[:,1])
        return is_in

    def equally_spaced_points(self):
        """Returns a grid of points within the region
        
        Returns
        -------
        all_points : np.ndarray of shape (num_points, 3) or (num_points, 2)
            The number of points returned is determined by the radius, height and point_spacing.
            Each row is a coordinate in 3D or 2D space.
        """
        num_points = np.floor(self.side_lengths / self.point_spacing)
        assert min(num_points) >= 1

        single_axes = [np.arange(num_p) * p_dist for num_p, p_dist in zip(num_points, self.point_spacing)]
        remainders = [s_len - single_ax[-1] for single_ax, s_len in zip(single_axes, self.side_lengths)]
        single_axes = [single_ax + remainder / 2 for single_ax, remainder in zip(single_axes, remainders)]
        all_points = np.meshgrid(*single_axes)
        all_points = np.concatenate([points.flatten()[:, None] for points in all_points], axis=-1)
        all_points = all_points + self.low_lim
        if self.spatial_dim == 3:
            all_points = np.concatenate((all_points, np.full((all_points.shape[0],1), self.center[-1])), axis=-1)
        return all_points

    def sample_points(self, num_points):
        """Returns a set of points sampled uniformly within the region
        
        Parameters
        ----------
        num_points : int
            Number of points to sample.

        Returns
        -------
        points : np.ndarray of shape (num_points, 3) or (num_points, 2)
            Each row is a coordinate in 3D or 2D space.
        """
        x = self.rng.uniform(self.low_lim[0], self.high_lim[0], num_points)
        y = self.rng.uniform(self.low_lim[1], self.high_lim[1], num_points)
        points = np.stack((x,y), axis=1)

        if self.spatial_dim == 3:
            points = np.concatenate((points, self.center[2]*np.ones((points.shape[0],1))), axis=-1)
        return points

    def plot(self, ax, label=None):
        rect = patches.Rectangle(self.low_lim, self.side_lengths[0], self.side_lengths[1], fill=True, alpha=0.3, label=label)
        ax.add_patch(rect)



class Disc(Region):
    def __init__(self, radius, center, point_spacing=(1,1), spatial_dim=3, rng=None):
        super().__init__(rng)
        assert spatial_dim in (2,3)
        assert len(center) == spatial_dim
        self.radius = radius
        self.center = np.array(center)
        self.point_spacing = np.array(point_spacing)
        self.spatial_dim = spatial_dim
        self.volume = self.radius**2 * np.pi

    def is_in_region(self, coordinates):
        if self.spatial_dim == 3:
            if not np.allclose(coordinates[:,2], self.center[2]):
                return False

        centered_coords = coordinates[:,:2] - self.center[None,:2]
        norm_coords = np.sqrt(np.sum(np.square(centered_coords), axis=-1))
        is_in = norm_coords <= self.radius
        return is_in

    def equally_spaced_points(self):
        point_dist = self.point_spacing
        block_dims = np.array([self.radius*2, self.radius*2])
        num_points = np.ceil(block_dims / point_dist)
        single_axes = [np.arange(num_p) * p_dist for num_p, p_dist in zip(num_points, point_dist)]

        meshed_points = np.meshgrid(*single_axes)
        meshed_points = [mp.reshape(-1)[:,None] for mp in meshed_points]
        all_points = np.concatenate(meshed_points, axis=-1)#.reshape(-1,2)

        shift = (num_points-1)*point_dist / 2
        all_points -= shift[None,:]

        inside_disc = np.sqrt(np.sum(all_points**2,axis=-1)) <= self.radius
        all_points = all_points[inside_disc,:]

        if self.spatial_dim == 3:
            all_points = np.concatenate((all_points, np.zeros((all_points.shape[0],1))), axis=-1)

        all_points += self.center[None,:]
        return all_points

    def sample_points(self, num_points):
        r = self.radius * np.sqrt(self.rng.uniform(0,1,num_points))
        angle = 2 * np.pi * self.rng.uniform(0,1,num_points)
        x = r * np.cos(angle) + self.center[0]
        y = r * np.sin(angle) + self.center[1]
        points = np.stack((x,y), axis=1)

        if self.spatial_dim == 3:
            points = np.concatenate((points, self.center[2]*np.ones((points.shape[0],1))), axis=-1)
        return points

    def plot(self, ax, label=None):
        circ = patches.Circle(self.center[:2], self.radius, fill=True, alpha=0.3, label=label)
        ax.add_patch(circ)



class Ball(Region):
    def __init__(self, radius, center, point_spacing=(1,1,1), rng=None):
        """Constructs a ball region
        
        Parameters
        ----------
        radius : float
            Radius of the ball.
        center : array_like of shape (3,), optional
            Center of the cylinder. The default is (0,0,0).
        point_spacing : array_like of shape (3,), optional
            Spacing between points in each direction, affects the selection of points
            for the equally_spaced_points method. The default is (1,1,1).
        rng : np.random.Generator, optional
            Random number generator. The default is None, in which case a new generator
            with a random seed will be created. 
            The generator affects the sampling of points in the sample_points method, and so 
            should be supplied for a reproducible result.

        Returns
        -------
        ball : Ball
        """
        super().__init__(rng)
        self.radius = radius
        self.center = np.array(center)
        assert self.center.ndim == 1
        self.point_spacing = np.array(point_spacing)

        self.volume = (4/3) * self.radius**3 * np.pi

    def is_in_region(self, coordinates):
        centered_coords = coordinates - self.center[None,:]
        is_in = np.linalg.norm(centered_coords, axis=-1) <= self.radius
        #is_in = norm_coords <= self.radius
        return is_in

    def equally_spaced_points(self):
        cuboid = Cuboid((2*self.radius, 2*self.radius, 2*self.radius), point_spacing=self.point_spacing)
        grid_points = cuboid.equally_spaced_points()

        grid_points += self.center[None,:]
        #inside_ball = np.linalg.norm(grid_points, axis=-1) <= self.radius
        grid_points = grid_points[self.is_in_region(grid_points),:]
        #grid_points += self.center[None,:]
        return grid_points

    def sample_points(self, num_points):
        finished = False
        num_accepted = 0

        samples = np.zeros((num_points, 3))
        while not finished:
            uniform_samples = self.rng.uniform(low=-self.radius, high=self.radius, size=(num_points, 3))

            filtered_samples = uniform_samples[np.linalg.norm(uniform_samples, axis=-1) <= self.radius,:]
            num_new = filtered_samples.shape[0]
            num_to_accept = min(num_new, num_points - num_accepted)

            samples[num_accepted:num_accepted+num_to_accept,:] = filtered_samples[:num_to_accept,:] 
            num_accepted += num_to_accept

            if num_accepted >= num_points:
                finished = True

        samples += self.center
        return samples

    def plot(self, ax, label=None):
        circ = patches.Circle(self.center[:2], self.radius, fill=True, alpha=0.3, label=label)
        ax.add_patch(circ)



class Cylinder(Region):
    def __init__(self, radius, height, center=(0,0,0), point_spacing=(1,1,1), rng=None):
        """Constructs a cylinder region
        
        Parameters
        ----------
        radius : float
            Radius of the cylinder.
        height : float
            Height of the cylinder.
        center : array_like of shape (3,), optional
            Center of the cylinder. The default is (0,0,0).
        point_spacing : array_like of shape (3,), optional
            Spacing between points in each direction, affects the selection of points
            for the equally_spaced_points method. The default is (1,1,1).
        rng : np.random.Generator, optional
            Random number generator. The default is None, in which case a new generator
            with a random seed will be created. 
            The generator affects the sampling of points in the sample_points method, and so 
            should be supplied for a reproducible result.

        Returns
        -------
        cylinder : Cylinder
        """
        super().__init__(rng)
        self.radius = radius
        self.height = height
        self.center = np.array(center)
        self.point_spacing = np.array(point_spacing)
        self.volume = self.radius**2 * np.pi * self.height

    def is_in_region(self, coordinates):
        """Checks whether the coordinate is within the cylinder or not. 
        
        Parameters
        ----------
        coordinates : np.ndarray
            Shape (N,3) where N is the number of coordinates to check. 
            Each row is a coordinate in 3D space.
            If only one coordinate is to be checked, it can be supplied as
            a (3,) array        

        Returns
        -------
        is_in : boolan np.ndarray of shape (N,)
            True if the coordinate is within the cylinder, False otherwise.
        """
        if coordinates.ndim == 1:
            coordinates = coordinates[None,:]
        assert coordinates.ndim == 2
        assert coordinates.shape[1] == 3

        centered_coords = coordinates - self.center[None,:]
        norm_coords = np.sqrt(np.sum(np.square(centered_coords[:,:2]), axis=-1))
        is_in_disc = norm_coords <= self.radius
        is_in_height = np.logical_and(-self.height/2 <= centered_coords[:,2],
                                      self.height/2 >= centered_coords[:,2])
        is_in = np.logical_and(is_in_height, is_in_disc)
        return is_in

    def equally_spaced_points(self):
        """Returns a grid of points within the cylinder
        
        Returns
        -------
        all_points : np.ndarray of shape (num_points, 3)
            The number of points returned is determined by the radius, height and point_spacing.
            Each row is a coordinate in 3D space.
        """
        point_dist = self.point_spacing
        block_dims = np.array([self.radius*2, self.radius*2, self.height])
        num_points = np.ceil(block_dims / point_dist)
        single_axes = [np.arange(num_p) * p_dist for num_p, p_dist in zip(num_points, point_dist)]
        
        all_points = np.stack(np.meshgrid(*single_axes, indexing="ij"),axis=-1)
        all_points = all_points.reshape(-1,3)

        shift = (num_points-1)*point_dist / 2
        all_points -= shift[None,:]

        inside_cylinder = np.sqrt(np.sum(all_points[:,:2]**2,axis=-1)) <= self.radius
        all_points = all_points[inside_cylinder,:]

        all_points += self.center[None,:]
        return all_points

    def sample_points(self, num_points):
        """Returns a set of points sampled uniformly within the cylinder
        
        Parameters
        ----------
        num_points : int
            Number of points to sample.

        Returns
        -------
        points : np.ndarray of shape (num_points, 3)
            Each row is a coordinate in 3D space.
        """
        r = self.radius * np.sqrt(self.rng.uniform(0,1,num_points))
        angle = 2 * np.pi * self.rng.uniform(0,1,num_points)
        x = r * np.cos(angle) + self.center[0]
        y = r * np.sin(angle) + self.center[1]
        h = self.rng.uniform(-self.height/2,self.height/2, num_points) + self.center[2]
        return np.stack((x,y,h), axis=1)

    def plot(self, ax, label=None):
        circ = patches.Circle(self.center[:2], self.radius, fill=True, alpha=0.3, label=label)
        ax.add_patch(circ)



# from abc import abstractmethod, ABC
# import numpy as np
# import itertools as it
# import ancsim.utilities as util

# class Shape(ABC):
#     def __init__(self):
#         self.area = 0
#         self.volume = 0

#     @abstractmethod
#     def draw(self, ax):
#         pass

#     @abstractmethod
#     def get_point_generator(self):
#         pass

#     @abstractmethod
#     def equivalently_spaced(self, num_points):
#         pass
    

# class Cuboid(Shape):
#     def __init__(self, size, center=(0,0,0)):
#         assert(len(size) == 3)
#         assert(len(center) == 3)
#         self.size = size
#         self.center = center
#         self.area = 2* (np.product(size[0:2]) + 
#                         np.product(size[1:3]) + 
#                         np.product(size[2:4]))
#         self.volume = np.product(size)

#         self._low_lim = center - (size / 2)
#         self._upper_lim = center + (size / 2)

#     def equivalently_spaced(self, grid_space):
#         n_points = self.size / grid_space
#         full_points = np.floor(n_points)
#         frac_points = n_points - full_points

#         np.linspace(self._low_lim[0], self._upper_lim[0], 2*n_points[0]+1)[1::2]

#     # def equivalently_spaced(self, num_points):
#     #     #pointsPerAxis = int(np.sqrt(numPoints/zNumPoints))
#     #     #p_distance = 0.05
#     #     #per_axis = self.size / p_distance

#     #     if self.size[0] == self.size[1]:
#     #         z_point_ratio = self.size[2] / self.size[0]
#     #         if util.isInteger(z_point_ratio):
                

#     #             self.size

#     #             #assert(np.isclose(pointsPerAxis**2*zNumPoints, numPoints))
#     #             x = np.linspace(self._low_lim[0], self._upper_lim[0], 2*pointsPerAxis+1)[1::2]
#     #             y = np.linspace(-dims[1]/2, dims[1]/2, 2*pointsPerAxis+1)[1::2]
#     #             z = np.linspace(-dims[2]/2, dims[2]/2, 2*zNumPoints+1)[1::2]
#     #             [xGrid, yGrid, zGrid] = np.meshgrid(x, y, z)
#     #             evalPoints = np.vstack((xGrid.flatten(), yGrid.flatten(), zGrid.flatten())).T

#     #             return evalPoints
#     #         else:
#     #             raise NotImplementedError
#     #     else:
#     #         raise NotImplementedError


        
#     def get_point_generator(self):
#         rng = np.random.RandomState(0)
#         def gen(num_samples):
#             points = np.vstack((rng.uniform(self._low_lim[0], self._upper_lim[0], num_samples)
#                                 rng.uniform(self._low_lim[1], self._upper_lim[1], num_samples)
#                                 rng.uniform(self._low_lim[2], self._upper_lim[2]], num_samples)))
#             return points
#         return gen



# class Cylinder(Shape):
#     def __init__(self, radius, height, center=(0,0,0)):
#         self.radius = radius
#         self.height = height
#         self.center = center
#         self.area = 2 * np.pi * (radius*height + radius**2)
#         self.volume = height * radius**2 * np.pi
        

#     def get_point_generator(self):
#         rng = np.random.RandomState(0)
#         def gen(numSamples):
#             r = np.sqrt(rng.rand(numSamples))*radius
#             angle = rng.rand(numSamples)*2*np.pi
#             x = r * np.cos(angle)
#             y = r * np.sin(angle)
#             z = rng.uniform(-height/2,height/2, size=numSamples)
#             points = np.stack((x,y,z))
#             return points
#         return gen



import numpy as np
import matplotlib.pyplot as plt

import aspsim.diagnostics.plot as dplt
import aspsim.array as ar

def compare_soundfields(pos, sig, algo_labels, row_labels, arrays_to_plot={}):
    """single pos array of shape (numPoints, spatialDim)
        sig and labels are lists of the same length, numAlgos
        each entry in sig is an array of shape (numAxes, numPoints), 
        so for example multiple frequencies can be plotted at the same time.
        
        The resulting plot is multiple axes, numAlgos rows high, and numAxes columns wide"""
    #num_rows, num_cols = sfp.get_num_pixels(pos)
    pos, sig = sort_for_imshow(pos, sig)
    if sig.ndim == 2:
        sig = sig[None,...]
    if sig.ndim == 3:
        sig = sig[None,...]

    num_algos = sig.shape[0]
    num_each_algo = sig.shape[1]
    extent = size_of_2d_plot([pos] + list(arrays_to_plot.values()))
    extent /= np.max(extent)
    scaling = 5

    fig, axes = plt.subplots(num_algos, num_each_algo, figsize=(num_each_algo*extent[0]*scaling, num_algos*extent[1]*scaling))
    fig.tight_layout(pad=2.5)

    


    for i, rows in enumerate(np.atleast_2d(axes)):
        for j, ax in enumerate(rows):
            sf_plot(ax, pos, sig[i,j,:,:], f"{algo_labels[i]}, {row_labels[j]}", arrays_to_plot)
    

def size_of_2d_plot(array_pos):
    """array_pos is a list/tuple of ndarrays of shape(any, spatialDim)
        describing the positions of all objects to be plotted. First axis 
        can be any value for the arrays, second axis must be 2 or more. Only
        the first two are considered. 
        
        returns np.array([x_size, y_size])"""
    array_pos = [ap[...,:2].reshape(-1,2) for ap in array_pos]
    all_pos = np.concatenate(array_pos, axis=0)
    extent = np.max(all_pos, axis=0) - np.min(all_pos, axis=0)  
    return extent[:2]
        

def sf_plot(ax, pos, sig, title="", arrays_to_plot = None, vminmax=None):
    """
        Makes a soundfield plot using the supplied matplotlib.pyplot axis

        Parameters
        ---------
        ax : axis obtained from fig, axes = plt.subplots()
        pos : ndarray of shape (num_pos, spatial_dim)
            must be sorted, use function sort_for_imshow()
        sig : ndarray of shape (num_pos, signal_dim)
            signal value for each position of the sound field
            must also be sorted the same way as pos
        title : str
            optional title for the plot
        arrays_to_plot : Array, ArrayCollection, list of Arrays, or dict of Arrays
            will be plotted along with the soundfield image. Can be used to see
            where microphones or loudspeakers are in relation to the soundfield
        vminmax : length-2 tuple of float
            the tuple should be (vmin, vmax), which sets the minimum and maximum value
            that the colors get assigned to. Use this in particular when plotting 
            multiple soundfields next to each other, so that the colors mean the
            same values in the different plots
        """

    if vminmax is None:
        vminmax = (np.min(sig), np.max(sig))

    im = ax.imshow(sig, interpolation="none", 
                    extent=(pos[...,0].min(), pos[...,0].max(), pos[...,1].min(), pos[...,1].max()), 
                    vmin=vminmax[0], vmax=vminmax[1],
                    cmap="magma",)
    
    if arrays_to_plot is not None:
        if isinstance(arrays_to_plot, ar.Array):
            arrays_to_plot = [arrays_to_plot]
        elif isinstance(arrays_to_plot, dict):
            arrays_to_plot = arrays_to_plot.values()
        for plot_ar in arrays_to_plot:
            plot_ar.plot(ax)
            #for arName, array in arrays_to_plot.items():
            #    ax.plot(array[:,0], array[:,1], "x", label=arName)

    ax.legend()
    ax.set_title(title)
    dplt.set_basic_plot_look(ax)
    ax.axis("equal")
    plt.colorbar(im, ax=ax, orientation='vertical')
    #plt.colorbar(ax=ax)

def get_num_pixels(pos, pos_decimals=5):
    pos_cols = np.unique(pos[:,0].round(pos_decimals))
    pos_rows = np.unique(pos[:,1].round(pos_decimals))
    num_rows = len(pos_rows)
    num_cols = len(pos_cols)
    return num_rows, num_cols


def sort_for_imshow(pos, sig, pos_decimals=5):
    """
        Sorts the position and signal values to display correctly when
        imshow is used to plot the sound field image

        Parameters
        ---------
        pos : ndarray of shape (num_pos, spatial_dim)
            must represent a rectangular grid, but can be in any order.
        sig : ndarray of shape (num_pos, signal_dim)
            signal value for each position of the sound field
        pos_decimals : int
            selects how many decimals the position values are rounded to when 
            calculating all the unique position values
        
        Returns
        -------
        pos_sorted : ndarray of shape (num_rows, num_cols, spatial_dim)
        sig_sorted : ndarray of shape (num_rows, num_cols, signal_dim)
    """
    if pos.shape[1] == 3:
        assert np.allclose(pos[:,2], np.ones_like(pos[:,2])*pos[0,2])

    num_rows, num_cols = get_num_pixels(pos, pos_decimals)
    unique_x = np.unique(pos[:,0].round(pos_decimals))
    unique_y = np.unique(pos[:,1].round(pos_decimals))

    sort_indices = np.zeros((num_rows, num_cols), dtype=int)
    for i, y in enumerate(unique_y):
        row_indices = np.where(np.abs(pos[:,1] - y) < 10**(-pos_decimals))[0]
        row_permutation = np.argsort(pos[row_indices,0])
        sort_indices[i,:] = row_indices[row_permutation]

    pos = pos[sort_indices,:]

    #sig = np.moveaxis(np.atleast_3d(sig),1,2)
    #dims = sig.shape[:2]
    signal_dim = sig.shape[-1]
    sig_sorted = np.zeros((num_rows, num_cols, signal_dim), dtype=sig.dtype)
    sig_sorted = np.flip(sig[sort_indices,:], axis=0)
    #for i in range(dims[0]):
     #   for j in range(dims[1]):
     #       single_sig = sig[i,j,:]
     #       sig_sorted[i,j,:,:] = np.flip(single_sig[sort_indices],axis=0)
    # sig_sorted = np.squeeze(sig_sorted)
    
    #sig = [np.flip(s[sortIndices],axis=0) for s in sig]
    
    return pos, sig_sorted






# def sort_for_imshow(pos, sig, pos_decimals=5):
#     """ 
#         Parameters
#         ---------
#         pos : ndarray of shape (num_pos, spatial_dim)
#         sig : ndarray of shape (num_pos, signal_dim)
#         pos_decimals : int
#             selects how many decimals the position values are rounded to when 
#             calculating all the unique position values

#         pos must be of shape (numPos, spatialDims) placed on a rectangular grid, 
#         but can be in any order.
#         sig can be a single signal or a list of signals of shape (numPos, signalDim), where each
#         entry on first axis is the value for pos[0,:] 
        
#         Returns
#         -------
#         pos_sorted : ndarray of shape (num_rows, num_cols, spatial_dim)
#         sig_sorted : ndarray of shape (num_rows, num_cols, signal_dim)
#     """
#     if pos.shape[1] == 3:
#         assert np.allclose(pos[:,2], np.ones_like(pos[:,2])*pos[0,2])

#     num_rows, num_cols = get_num_pixels(pos, pos_decimals)
#     unique_x = np.unique(pos[:,0].round(pos_decimals))
#     unique_y = np.unique(pos[:,1].round(pos_decimals))

#     sort_indices = np.zeros((num_rows, num_cols), dtype=int)
#     for i, y in enumerate(unique_y):
#         row_indices = np.where(np.abs(pos[:,1] - y) < 10**(-pos_decimals))[0]
#         row_permutation = np.argsort(pos[row_indices,0])
#         sort_indices[i,:] = row_indices[row_permutation]

#     pos = pos[sort_indices,:]

#     sig = np.moveaxis(np.atleast_3d(sig),1,2)
#     dims = sig.shape[:2]
#     sig_sorted = np.zeros((dims[0], dims[1], num_rows, num_cols), dtype=sig.dtype)
#     for i in range(dims[0]):
#         for j in range(dims[1]):
#             single_sig = sig[i,j,:]
#             sig_sorted[i,j,:,:] = np.flip(single_sig[sort_indices],axis=0)
#     sig_sorted = np.squeeze(sig_sorted)
    
#     #sig = [np.flip(s[sortIndices],axis=0) for s in sig]
    
#     return pos, sig_sorted


import aspsim.room.generatepoints as gp

import numpy as np
import matplotlib.pyplot as plt







def test_show_equidistance_rectangle():

    num_points = 64
    points = gp.equidistant_rectangle(num_points, (2,1), extra_side_lengths=(0.5, 0), offset=0.5, z = 0)

    fig, ax  = plt.subplots()
    for m in range(num_points):
        ax.text(points[m,0], points[m,1], str(m))
    plt.plot(points[:,0], points[:,1], "x")
    plt.show()



import numpy as np

import aspsim.room.generatepoints as gp

class Trajectory:
    """Base class defining a trajectory through the room. 
    
    In order to simulate a moving source or microphone, the trajectory must be 
    defined in accordance with this API. Easiest way is to create a function defining 
    the position at each time index, and pass it to the constructor. Alternatively, 
    subclass this class and implement the current_pos method.
    """
    def __init__(self, pos_func):
        """pos_func is a function, which takes a time_index in samples and outputs a position"""
        self.pos_func = pos_func
        #self.pos = np.full((1,3), np.nan)

    def current_pos(self, time_idx):
        """
        Parameters
        ----------
        time_idx : int
            time index in samples

        Returns
        -------
        pos : array of shape (1,3)
            position at time time_idx
        """
        return self.pos_func(time_idx)

    def plot(self, ax, symbol, name, tot_samples):
        pass


class TrajectoryCollection(Trajectory):
    """A class for combining multiple trajectories into one, in the case where you want to have multiple moving objects in the same array. 
    """
    def __init__(self, trajectories):
        """
        Parameters
        ----------
        trajectories : list of Trajectory objects
        """
        self.trajectories = trajectories
        #self.num_pos = len(self.trajectories)

    def current_pos(self, time_idx):
        return np.concatenate([traj.current_pos(time_idx) for traj in self.trajectories], axis=0)
    
    def plot(self, ax, symbol, name, tot_samples):
        for traj in self.trajectories:
            traj.plot(ax, symbol, name, tot_samples)

class LinearTrajectory(Trajectory):
    def __init__(self, points, period, samplerate, mode="constant_speed"):
        """A trajectory that moves through a series of points in straight lines.

        Parameters
        ----------
        points : ndarray of shape (numpoints, spatial_dim) or equivalent list of lists
            The points that the trajectory will move through. The trajectory will
            start and end at the first point.
        period : float
            The time in seconds for the trajectory to go through all the points and
            return to the starting point.
        samplerate : int
            The samplerate of the simulation. 
        mode : 'constant_speed' or 'constant_time'
            if 'constant_speed', the speed of the movement will be constant, and
            calibrated such that it returns to the starting position after one period.
            if 'constant_time', each segment will take equal time, and the speed will
            therefore go up for long segments and down for short segments.
        """
        if isinstance(points, (list, tuple)):
            points = np.array(points)
        #self.num_pos = 1

        if not np.allclose(points[-1,:], points[0,:]):
            points = np.concatenate((points, points[:1,:]), axis=0)
        self.anchor_points = points
        self.period = period
        self.samplerate = samplerate

        if mode == "constant_speed":
            pos_func = self._constant_speed_pos_func(points, period, samplerate)
        elif mode == "constant_time":
            pos_func = self._constant_time_pos_func(points, period, samplerate)
        else:
            raise ValueError ("Invalid mode argument")

        super().__init__(pos_func)

    def _constant_speed_pos_func(self, points, period, samplerate):
        segment_distance = np.sqrt(np.sum((points[:-1,:] - points[1:,:])**2, axis=-1))
        assert all(segment_distance > 0)
        
        tot_distance = np.sum(segment_distance)
        distance_per_sample = tot_distance / (samplerate * period)
        self.samples_per_segment = segment_distance / distance_per_sample

        #cycle_start = 0
        self.cumulative_samples = np.concatenate(([0], np.cumsum(self.samples_per_segment)))

        def pos_func(n):
            period_samples = n % (period * samplerate)

            point_idx = np.argmax(period_samples < self.cumulative_samples)
            time_this_segment = period_samples - self.cumulative_samples[point_idx-1]
            ip_factor = time_this_segment / self.samples_per_segment[point_idx-1]
            pos = points[point_idx-1,:]*(1-ip_factor) + points[point_idx,:]*ip_factor
            return pos[None,:]
        return pos_func
    
    def _constant_time_pos_func(self, points, period, samplerate):
        num_segments = points.shape[0]
        time_per_segment = period / num_segments
        self.samples_per_segment = time_per_segment * samplerate
        segment_distance = np.sqrt(np.sum((points[:-1,:] - points[1:,:])**2, axis=-1))
        
        distance_per_time = segment_distance / time_per_segment
        distance_per_sample = distance_per_time / samplerate

        self.cumulative_samples = np.arange(num_segments) * self.samples_per_segment #np.concatenate(([0], np.cumsum(samples_per_segment)))

        def pos_func(n):
            period_samples = n % (period * samplerate)

            point_idx = np.argmax(period_samples < self.cumulative_samples)
            time_this_segment = period_samples - self.cumulative_samples[point_idx-1] # time in number of samples
            ip_factor = time_this_segment / self.samples_per_segment
            pos = points[point_idx-1,:]*(1-ip_factor) + points[point_idx,:]*ip_factor
            return pos[None,:]
        return pos_func

    def plot(self, ax, symbol, label, tot_samples=None):
        if tot_samples is None:
            points = self.anchor_points
        else:
            if tot_samples >= self.period*self.samplerate:
                points = self.anchor_points
            else:
                point_idx = np.argmax(tot_samples < self.cumulative_samples)
                points = self.anchor_points[:point_idx,:]
        ax.plot(points[:,0], points[:,1], marker=symbol, linestyle="dashed", label=label, alpha=0.8)

    # @classmethod
    # def linear_interpolation_const_speed(cls, points, period, samplerate):
    #     """
    #     points is array of shape (numpoints, spatial_dim) or equivalent list of lists
    #     period is in seconds
    #     update freq is in samples
    #     """
    #     if isinstance(points, (list, tuple)):
    #         points = np.array(points)

    #     if not np.allclose(points[-1,:], points[0,:]):
    #         points = np.concatenate((points, points[:1,:]), axis=0)

    #     segment_distance = np.sqrt(np.sum((points[:-1,:] - points[1:,:])**2, axis=-1))
    #     assert all(segment_distance > 0)
        
    #     tot_distance = np.sum(segment_distance)
    #     #updates_per_period = samplerate * period / update_freq
    #     distance_per_sample = tot_distance / (samplerate * period)
    #     segment_samples = segment_distance / distance_per_sample

    #     cycle_start = 0
    #     cumulative_samples = np.concatenate(([0], np.cumsum(segment_samples)))

    #     def pos_func(t):
    #         period_samples = t % (period * samplerate)

    #         point_idx = np.argmax(period_samples < cumulative_samples)
    #         time_this_segment = period_samples - cumulative_samples[point_idx-1]
    #         ip_factor = time_this_segment / segment_samples[point_idx-1]
    #         pos = points[point_idx-1,:]*(1-ip_factor) + points[point_idx,:]*ip_factor
    #         return pos[None,:]

    #     return cls(pos_func)




class CircularTrajectory(Trajectory):
    def __init__(self, 
            radius : tuple[float, float], 
            center : tuple[float, float, float], 
            radial_period : float, 
            angle_period : float, 
            samplerate : int,
            start_angle : float = 0, 
            ):
        """A trajectory that moves around a circle. 

        Moves around a circle in one angle_period, while it moves from the outer radius
        to the inner radius and back again in one radial_period

        The circle is defined by its center and radius

        Parameters
        ----------
        radius : length-2 tuple of floats
            inner and outer radius. Interpreted by aspsim as meters
        center : length-3 tuple of floats
            center of the disc/circle
        radial_period : float
            time in seconds for the trajectory to go from outer radius to 
            inner radius and back again
        angle_period : float
            time in seconds for trajectory to go one lap around the circle
        samplerate : int
            samplerate of the simulation, supplied for the units of the periods 
            to make sense
        
        """
        self.radius = radius
        self.center = center
        self.radial_period = radial_period
        self.angle_period = angle_period
        self.samplerate = samplerate
        self.start_angle = start_angle
        #self.num_pos = 1

        self.radius_diff = self.radius[1] - self.radius[0]
        assert self.radius_diff >= 0

        def pos_func(t):
            angle_period_samples = t % (angle_period * samplerate)
            radial_period_samples = t % (radial_period * samplerate)

            angle_portion = angle_period_samples / (angle_period * samplerate)
            radial_portion = radial_period_samples / (radial_period * samplerate)

            angle = self.start_angle + 2 * np.pi * angle_portion
            #angle = angle % (2*np.pi)

            if radial_portion < 0.5:
                rad = radius[0] + self.radius_diff * (1 - 2 * radial_portion)
            else:
                rad = radius[0] + self.radius_diff * (radial_portion-0.5)*2

            (x,y) = gp.pol2cart(rad, angle)
            return np.array([[x,y,0]]) + center

        super().__init__(pos_func)

    def plot(self, ax, symbol="o", label="", tot_samples=None):
        #if tot_samples is not None:
        #    max_samples = 
        #    raise NotImplementedError
        
        approx_num_points = 1000
        max_samples = self.samplerate*max(self.radial_period, self.angle_period)
        samples_per_point = max_samples // approx_num_points
        t = np.arange(0, max_samples, samples_per_point)
        num_points = t.shape[0]

        pos = np.zeros((num_points, 3))
        for i in range(num_points):
            pos[i,:] = np.squeeze(self.pos_func(t[i]))
        ax.plot(pos[:,0], pos[:,1], marker=symbol, linestyle="dashed", label=label, alpha=0.8)

        # ax.plot(self.anchor_points[:,0], self.anchor_points[:,1], marker=symbol, linestyle="dashed", label=label, alpha=0.8)



import aspsim.room.region as region
import numpy as np

def _random_cylinder(rng=None):
    if rng is None:
        rng = np.random.default_rng()
    radius = rng.uniform(0.5, 2)
    height = rng.uniform(0.5, 2)
    center = rng.uniform(-1, 1, size=3)
    point_spacing = rng.uniform(0.1, 0.5, size=3)
    cylinder = region.Cylinder(radius, height, center, point_spacing, rng=rng)
    return cylinder

def _random_cuboid(rng=None):
    if rng is None:
        rng = np.random.default_rng()
    side_lengths = rng.uniform(0.5, 2, size=3)
    center = rng.uniform(-1, 1, size=3)
    point_spacing = rng.uniform(0.1, 0.5, size=3)
    cuboid = region.Cuboid(side_lengths, center, point_spacing, rng=rng)
    return cuboid

def _random_ball(rng=None):
    if rng is None:
        rng = np.random.default_rng()
    radius = rng.uniform(0.5, 2)
    center = rng.uniform(-1, 1, size=3)
    point_spacing = rng.uniform(0.1, 0.5, size=3)
    ball = region.Ball(radius, center, point_spacing, rng=rng)
    return ball


def _random_rectangle(rng=None):
    if rng is None:
        rng = np.random.default_rng()
    side_lengths = rng.uniform(0.5, 2, size=2)
    center = rng.uniform(-1, 1, size=3)
    point_spacing = rng.uniform(0.1, 0.5, size=2)
    rect = region.Rectangle(side_lengths, center, point_spacing, rng=rng)
    return rect

def test_cylinder_equally_spaced_points_returns_points_within_region():
    cyl = _random_cylinder()
    points = cyl.equally_spaced_points()

    for i in range(points.shape[0]):
        point = points[i,:]
        assert (point[0]-cyl.center[0]) ** 2 + (point[1]-cyl.center[1]) ** 2 <= cyl.radius ** 2
        assert cyl.center[2] - cyl.height / 2 <= point[2] <= cyl.center[2] + cyl.height / 2
        assert cyl.is_in_region(point[None,:])

def test_cylinder_sample_points_returns_points_within_region():
    cyl = _random_cylinder()
    num_to_sample = 1000
    points = cyl.sample_points(num_to_sample)

    for i in range(points.shape[0]):
        point = points[i,:]
        assert (point[0]-cyl.center[0]) ** 2 + (point[1]-cyl.center[1]) ** 2 <= cyl.radius ** 2
        assert cyl.center[2] - cyl.height / 2 <= point[2] <= cyl.center[2] + cyl.height / 2
        assert cyl.is_in_region(point[None,:])


def test_cylinder_a_random_point_is_always_close_to_equally_spaced_point():
    cyl = _random_cylinder()
    points = cyl.equally_spaced_points()

    num_to_sample = 1000
    points_random = cyl.sample_points(num_to_sample)

    for i in range(points_random.shape[0]):
        diff = points - points_random[i:i+1,:]
        min_distance = np.min(np.abs(diff), axis=0)
        assert np.all(min_distance < cyl.point_spacing)



def test_cuboid_equally_spaced_points_returns_points_within_region():
    cb = _random_cuboid()
    points = cb.equally_spaced_points()

    for i in range(points.shape[0]):
        point = points[i,:]
        assert np.all(cb.center - cb.side_lengths / 2 <= point)
        assert np.all(point <= cb.center + cb.side_lengths / 2)
        assert cb.is_in_region(point[None,:])

def test_cuboid_sample_points_returns_points_within_region():
    cb = _random_cuboid()
    num_to_sample = 1000
    points = cb.sample_points(num_to_sample)

    for i in range(points.shape[0]):
        point = points[i,:]
        assert np.all(cb.center - cb.side_lengths / 2 <= point)
        assert np.all(point <= cb.center + cb.side_lengths / 2)
        assert cb.is_in_region(point[None,:])

def test_cuboid_a_random_point_is_always_close_to_equally_spaced_point():
    cb = _random_cuboid()
    points = cb.equally_spaced_points()

    num_to_sample = 1000
    points_random = cb.sample_points(num_to_sample)

    for i in range(points_random.shape[0]):
        diff = points - points_random[i:i+1,:]
        min_distance = np.min(np.abs(diff), axis=0)
        assert np.all(min_distance < cb.point_spacing)




def test_ball_equally_spaced_points_returns_points_within_region():
    ball = _random_ball()
    points = ball.equally_spaced_points()

    for i in range(points.shape[0]):
        point = points[i,:]
        assert np.all(np.linalg.norm(point - ball.center, axis=-1) <= ball.radius) 
        assert ball.is_in_region(point[None,:])

def test_ball_sample_points_returns_points_within_region():
    ball = _random_ball()
    num_to_sample = 1000
    points = ball.sample_points(num_to_sample)

    for i in range(points.shape[0]):
        point = points[i,:]
        assert np.all(np.linalg.norm(point - ball.center, axis=-1) <= ball.radius) 
        assert ball.is_in_region(point[None,:])

def test_ball_a_random_point_is_always_close_to_equally_spaced_point():
    ball = _random_ball()
    points = ball.equally_spaced_points()

    num_to_sample = 1000
    points_random = ball.sample_points(num_to_sample)

    for i in range(points_random.shape[0]):
        diff = points - points_random[i:i+1,:]
        min_distance = np.min(np.abs(diff), axis=0)
        assert np.all(min_distance < ball.point_spacing)




def test_rectangle_equally_spaced_points_returns_points_within_region():
    rect = _random_rectangle()
    points = rect.equally_spaced_points()

    for i in range(points.shape[0]):
        point = points[i,:]
        side_len_3d = np.concatenate((rect.side_lengths, [0]))
        assert np.all(rect.center - side_len_3d / 2 <= point)
        assert np.all(point <= rect.center + side_len_3d / 2)
        assert rect.is_in_region(point[None,:])

def test_rectangle_sample_points_returns_points_within_region():
    rect = _random_rectangle()
    num_to_sample = 1000
    points = rect.sample_points(num_to_sample)

    for i in range(points.shape[0]):
        point = points[i,:]
        side_len_3d = np.concatenate((rect.side_lengths, [0]))
        assert np.all(rect.center - side_len_3d / 2 <= point)
        assert np.all(point <= rect.center + side_len_3d / 2)
        assert rect.is_in_region(point[None,:])

def test_rectangle_a_random_point_is_always_close_to_equally_spaced_point():
    rect = _random_rectangle()
    points = rect.equally_spaced_points()

    num_to_sample = 1000
    points_random = rect.sample_points(num_to_sample)

    for i in range(points_random.shape[0]):
        diff = points - points_random[i:i+1,:]
        min_distance = np.min(np.abs(diff), axis=0)
        assert np.all(min_distance[:2] < rect.point_spacing)


import numpy as np
import json
import scipy.signal as spsig

import aspsim.diagnostics.plot as dplt
import aspsim.diagnostics.core as diacore

def add_to_summary(diagName, summaryValues, timeIdx, folderPath):
    fullPath = folderPath.joinpath("summary_" + str(timeIdx) + ".json")
    try:
        with open(fullPath, "r") as f:
            summary = json.load(f)
            summary[diagName] = summaryValues
            # totData = {**oldData, **dictToAdd}
    except FileNotFoundError:
        summary = {}
        summary[diagName] = summaryValues
    with open(fullPath, "w") as f:
        json.dump(summary, f, indent=4)


def mean_near_time_idx(outputs, timeIdx):
    summaryValues = {}
    numToAverage = 3000

    for filtName, output in outputs.items():
        # val = diagnostic.getOutput()[timeIdx-numToAverage:timeIdx]
        val = output[...,timeIdx - numToAverage : timeIdx]
        filterArray = np.logical_not(np.isnan(val))
        summaryValues[filtName] = np.mean(val[filterArray])
    return summaryValues


def last_value():
    raise NotImplementedError






class SummaryDiagnostic(diacore.Diagnostic):
    export_functions = {
        "npz" : dplt.savenpz,
        "text" : dplt.txt,
        "spectrum" : dplt.spectrum_plot,
    }
    def __init__ (
        self,
        sim_info,
        block_size,
        save_at,
        export_func = "text",
        keep_only_last_export = False,
        export_kwargs = None,
        preprocess = None,
        ):
        """
        save_at should be a tuple (start_sample, end_sample)
            it will use the samples between start_samle (inclusive) and end_sample (exclusive)
        
        """
        if isinstance(save_at, diacore.IntervalCounter):
            raise NotImplementedError
        else:
            export_at = [save_at[1]]
            save_at = diacore.IntervalCounter(((save_at[0], save_at[1]),))

        super().__init__(sim_info, block_size, export_at, save_at, export_func, keep_only_last_export, export_kwargs, preprocess)






class SignalPowerRatioSummary(SummaryDiagnostic):
    def __init__(self, 
        numerator_name,
        denom_name,
        sim_info, 
        block_size, 
        save_range,
        numerator_channels=slice(None),
        denom_channels = slice(None),
        **kwargs
        ):
        self.save_range = save_range
        self.num_samples = save_range[1] - save_range[0]
        super().__init__(sim_info, block_size, save_range, **kwargs)
        self.numerator_name = numerator_name
        self.denom_name = denom_name
        self.num_power = 0
        self.denom_power = 0

        self.numerator_channels = numerator_channels
        self.denom_channels = denom_channels

        self.plot_data["title"] = f"Ratio of power: {self.numerator_name} / {self.denom_name}. Samples: {self.save_range}"
        
    def save(self, processor, sig, chunkInterval, globInterval):
        self.num_power += np.sum(np.mean(np.abs(processor.sig[self.numerator_name][self.numerator_channels, chunkInterval[0]:chunkInterval[1]])**2, axis=0)) / self.num_samples
        self.denom_power += np.sum(np.mean(np.abs(processor.sig[self.denom_name][self.denom_channels, chunkInterval[0]:chunkInterval[1]])**2,axis=0)) / self.num_samples

        #self.power_ratio[globInterval[0]:globInterval[1]] = num / denom

    def get_output(self):
        return self.num_power / self.denom_power


class SignalPowerSummary(SummaryDiagnostic):
    def __init__(self, 
        sig_name,
        sim_info, 
        block_size, 
        save_range,
        sig_channels=slice(None),
        **kwargs
        ):
        """
        Will output the wrong value if it is exported in the middle of the save_range
        """
        self.save_range = save_range
        self.num_samples = save_range[1] - save_range[0]
        super().__init__(sim_info, block_size, save_range, **kwargs)
        self.sig_name = sig_name
        self.power = 0

        self.sig_channels = sig_channels

        #self.plot_data["title"] = f"Power of {self.sig_name}. Samples: {self.save_range}"
        
    def save(self, processor, sig, chunkInterval, globInterval):
        self.power += np.sum(np.mean(np.abs(processor.sig[self.sig_name][self.sig_channels, chunkInterval[0]:chunkInterval[1]])**2, axis=0)) / self.num_samples

        #self.power_ratio[globInterval[0]:globInterval[1]] = num / denom

    def get_output(self):
        return self.power



class SignalPowerSpectrum(SummaryDiagnostic):
    def __init__(self, 
        sig_name,
        sim_info, 
        block_size, 
        save_range,
        num_channels,
        sig_channels=slice(None),
        **kwargs
        ):
        """
        Will output the wrong value if it is exported in the middle of the save_range
        """
        self.save_range = save_range
        self.num_samples = save_range[1] - save_range[0]
        super().__init__(sim_info, block_size, save_range, export_func = "spectrum", **kwargs)
        self.sig_name = sig_name
        self.power = 0

        self.samplerate = self.sim_info.samplerate
        self.num_channels = num_channels
        self.sig_channels = sig_channels
        self.power = np.full((self.num_channels, self.num_samples), fill_value=np.nan)

        self.sample_counter = 0

        #self.plot_data["title"] = f"Power of {self.sig_name}. Samples: {self.save_range}"
        
    def save(self, processor, sig, chunkInterval, globInterval):
        num_samples = chunkInterval[1] - chunkInterval[0]
        self.power[:,self.sample_counter:self.sample_counter+num_samples] = np.abs(processor.sig[self.sig_name][self.sig_channels, chunkInterval[0]:chunkInterval[1]])**2

        self.sample_counter += num_samples
        #self.power_ratio[globInterval[0]:globInterval[1]] = num / denom

    def get_output(self):
        f, spec = spsig.welch(self.power, self.samplerate, nperseg=512, scaling="spectrum", axis=-1)
        spec = np.mean(spec, axis=0)
        return spec



import numpy as np
from abc import ABC, abstractmethod

import aspcore.filter as fc

import aspsim.diagnostics.core as diacore
import aspsim.diagnostics.diagnostics as dia
from aspsim.simulator import Signals

class AudioProcessor(ABC):
    def __init__(self, sim_info, arrays, block_size, diagnostics={}, rng=None):
        self.sim_info = sim_info
        self.arrays = arrays
        self.block_size = block_size

        self.name = "Abstract Processor"
        self.metadata = {}
        self.sig = Signals(self.sim_info, self.arrays)

        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

    def prepare(self):
        pass

    @abstractmethod
    def process(self, num_samples):
        """ microphone signals up to self.idx (excluding) are available. i.e. [:,choose_start_idx:self.idx] can be used
            To play a signal through controllable loudspeakers, add the values to self.sig['name-of-loudspeaker']
            for samples self.idx and forward, i.e. [:,self.idx:self.idx+self.block_size]. 
            Adding samples further ahead than that will likely cause a outOfBounds error. 
        """
        pass



class DebugProcessor(AudioProcessor):
    def __init__(self, sim_info, arrays, block_size, **kwargs):
        super().__init__(sim_info, arrays, block_size, **kwargs)
        self.name = "Debug Processor"
        self.processed_samples = 0
        self.manual_idx = 0
        self.filt = fc.create_filter(num_in=3, num_out=4, ir_len=5)

        self.mic = np.zeros((self.arrays["mic"].num, self.sim_info.tot_samples))
        self.ls = np.zeros((self.arrays["loudspeaker"].num, self.sim_info.tot_samples))
        

    def process(self, num_samples):
        self.sig["loudspeaker"][:,self.sig.idx:self.sig.idx+num_samples] = \
            self.sig["mic"][:,self.sig.idx-num_samples:self.sig.idx]

        for manual_i, i in zip(range(self.manual_idx,self.manual_idx+num_samples),range(self.sig.idx-num_samples, self.sig.idx)):
            if manual_i < self.sim_info.tot_samples:
                self.mic[:,manual_i] = self.sig["mic"][:,i]
                self.ls[:,manual_i] = self.sig["loudspeaker"][:,i]
                self.manual_idx += 1

        self.processed_samples += num_samples
        self.filt.ir += num_samples
















def calc_block_sizes_with_buffer(num_samples, idx, buffer_size, chunk_size):
    leftInBuffer = chunk_size + buffer_size - idx
    sampleCounter = 0
    block_sizes = []
    while sampleCounter < num_samples:
        bLen = np.min((num_samples - sampleCounter, leftInBuffer))
        block_sizes.append(bLen)
        sampleCounter += bLen
        leftInBuffer -= bLen
        if leftInBuffer == 0:
            leftInBuffer = chunk_size
    return block_sizes


def find_first_index_for_block(earliest_start_index, index_to_end_at, block_size):
    """If processing in fixed size blocks, this function will give the index
        to start at if the processing should end at a specific index.
        Useful for preparation processing, where the exact startpoint isn't important
        but it is important to end at the correct place. """
    num_samples = index_to_end_at - earliest_start_index
    num_blocks = num_samples // block_size
    index_to_start_at = index_to_end_at - block_size*num_blocks
    return index_to_start_at

def block_process_until_index(earliest_start_index, index_to_end_at, block_size):
    """Use as 
        for startIdx, endIdx in blockProcessUntilIndex(earliestStart, indexToEnd, block_size):
            process(signal[...,startIdx:endIdx])
    
        If processing in fixed size blocks, this function will give the index
        to process for, if the processing should end at a specific index.
        Useful for preparation processing, where the exact startpoint isn't important
        but it is important to end at the correct place. """
    num_samples = index_to_end_at - earliest_start_index
    num_blocks = num_samples // block_size
    index_to_start_at = index_to_end_at - block_size*num_blocks

    for i in range(num_blocks):
        yield index_to_start_at+i*block_size, index_to_start_at+(i+1)*block_size




import numpy as np
import scipy.linalg as splin
import scipy.signal as spsig

import aspsim.diagnostics.core as diacore
import aspsim.diagnostics.plot as dplot


class EigenvaluesOverTime(diacore.StateDiagnostic):
    """
    Matrix must be square, otherwise EVD doesn't work
    For now assumes hermitian matrix as well

    eigval_idx should be a tuple with the indices of the desired eigenvalues
    ascending order, zero indexed, and top inclusive

    The first value of num_eigvals is how many of the lowest eigenvalues that should be recorded
    The seconds value is how many of the largest eigenvalues that should be recorded
    """
    def __init__(self, matrix_name, eigval_idx, *args, abs_value=False, **kwargs):
        super().__init__(*args, **kwargs)
        assert isinstance(eigval_idx, (list, tuple, np.ndarray))
        assert len(eigval_idx) == 2
        self.get_matrix = diacore.attritemgetter(matrix_name)
        self.eigval_idx = eigval_idx
        self.abs_value = abs_value

        self.eigvals = np.full((eigval_idx[1]-eigval_idx[0]+1, self.save_at.num_values), np.nan)
        self.time_indices = np.full((self.save_at.num_values), np.nan, dtype=int)
        self.diag_idx = 0

        if self.abs_value:
            self.plot_data["title"] = "Size of Eigenvalues"
        else:
            self.plot_data["title"] = "Eigenvalues"

    def save(self, processor, sig, chunkInterval, globInterval):
        assert globInterval[1] - globInterval[0] == 1
        mat = self.get_matrix(processor)
        assert np.allclose(mat, mat.T.conj())
        evs = splin.eigh(mat, eigvals_only=True, subset_by_index=self.eigval_idx)
        if self.abs_value:
            evs = np.abs(evs)
        self.eigvals[:, self.diag_idx] = evs
        self.time_indices[self.diag_idx] = globInterval[0]

        self.diag_idx += 1
        
    def get_output(self):
        return self.eigvals


class Eigenvalues(diacore.InstantDiagnostic):
    def __init__ (
        self, 
        matrix_name,
        *args,
        abs_value = False,
        **kwargs,
        ):
        super().__init__(*args, **kwargs)
        #self.matrix_name = matrix_name
        self.get_mat = diacore.attritemgetter(matrix_name)
        self.evs = None
        self.abs_value = abs_value

        if self.abs_value:
            self.plot_data["title"] = "Size of Eigenvalues"
        else:
            self.plot_data["title"] = "Eigenvalues"

    def save(self, processor, sig, chunkInterval, globInterval):
        mat = self.get_mat(processor)
        assert np.allclose(mat, mat.T.conj())
        self.evs = splin.eigh(mat, eigvals_only=True)[None,None,:]

        if self.abs_value:
            self.evs = np.abs(self.evs)

    def get_output(self):
        return self.evs





class SummaryDiagnostic(diacore.Diagnostic):
    export_functions = {
        "npz" : dplot.savenpz,
        "text" : dplot.txt,
        "spectrum" : dplot.spectrum_plot,
    }
    def __init__ (
        self,
        sim_info,
        save_at,
        export_func = "text",
        keep_only_last_export = False,
        export_kwargs = None,
        preprocess = None,
        ):
        """
        save_at should be a tuple (start_sample, end_sample)
            it will use the samples between start_samle (inclusive) and end_sample (exclusive)
        
        """
        if isinstance(save_at, diacore.IntervalCounter):
            raise NotImplementedError
        else:
            export_at = [save_at[1]]
            save_at = diacore.IntervalCounter(((save_at[0], save_at[1]),))

        super().__init__(sim_info, export_at, save_at, export_func, keep_only_last_export, export_kwargs, preprocess)


class SignalPowerSpectrum(SummaryDiagnostic):
    def __init__(self, 
        sig_name,
        sim_info, 
        save_range,
        num_channels,
        nperseg, 
        sig_channels=slice(None),
        **kwargs
        ):
        """
        Will output the wrong value if it is exported in the middle of the save_range
        """
        self.save_range = save_range
        self.num_samples = save_range[1] - save_range[0]
        #save_at = diacore.IntervalCounter((save_range,))
        super().__init__(sim_info, save_at=save_range, export_func = "spectrum", **kwargs)
        self.sig_name = sig_name

        self.samplerate = self.sim_info.samplerate
        self.num_channels = num_channels
        self.nperseg = nperseg
        self.sig_channels = sig_channels
        self.signal = np.full((self.num_channels, self.num_samples), fill_value=np.nan)

        self.sample_counter = 0

        

        #self.plot_data["title"] = f"Power of {self.sig_name}. Samples: {self.save_range}"
        
    def save(self, processor, sig, chunkInterval, globInterval):
        num_samples = chunkInterval[1] - chunkInterval[0]
        self.signal[:,self.sample_counter:self.sample_counter+num_samples] = sig[self.sig_name][self.sig_channels, chunkInterval[0]:chunkInterval[1]]

        self.sample_counter += num_samples
        #self.power_ratio[globInterval[0]:globInterval[1]] = num / denom

    def get_output(self):
        f, spec = spsig.welch(self.signal, self.samplerate, nperseg=self.nperseg, scaling="spectrum", axis=-1)
        spec = np.mean(spec, axis=0)
        return spec



class SignalSummary(SummaryDiagnostic):
    def __init__(self, 
        sig_name,
        sim_info, 
        save_range,
        summary_func = None,
        **kwargs
        ):
        self.save_range = save_range
        self.num_samples = save_range[1] - save_range[0]
        #save_at = diacore.IntervalCounter((save_range,))
        super().__init__(sim_info, save_at=save_range, export_func = "text", **kwargs)
        self.sig_name = sig_name
        self.summary_func = summary_func

        self.mean = 0

        #self.plot_data["title"] = f"Power of {self.sig_name}. Samples: {self.save_range}"
        
    def save(self, processor, sig, chunkInterval, globInterval):
        if self.summary_func is None:
            self.mean += np.sum(sig[self.sig_name][:, chunkInterval[0]:chunkInterval[1]]) / self.num_samples
        else:
            self.mean += self.summary_func(sig[self.sig_name][:, chunkInterval[0]:chunkInterval[1]]) / self.num_samples

    def get_output(self):
        return self.mean


import numpy as np
import copy
import aspcore.filter as fc
import aspsim.diagnostics.core as diacore
import aspsim.diagnostics.preprocessing as pp
import aspsim.diagnostics.plot as dplot

class RecordFilter(diacore.InstantDiagnostic):
    """
        Remember to include .ir in the property name
        if the property is a filter object
    """
    def __init__ (
        self, 
        prop_name, 
        *args,
        **kwargs,
        ):
        super().__init__(*args, **kwargs)
        self.prop_name = prop_name
        self.get_prop = diacore.attritemgetter(prop_name)
        self.prop = None

    def save(self, processor, sig, chunkInterval, globInterval):
        self.prop = copy.deepcopy(self.get_prop(processor))

    def get_output(self):
        return self.prop


class RecordFilterDifference(RecordFilter):
    """
    """
    def __init__(self, 
        state_name,
        state_name_subtract,
        *args,
        **kwargs):
        super().__init__(state_name, *args, **kwargs)
        self.get_state_subtract = diacore.attritemgetter(state_name_subtract)

    def save(self, processor, sig, chunkInterval, globInterval):
        self.prop = copy.deepcopy(self.get_prop(processor)) - copy.deepcopy(self.get_state_subtract(processor))



# Both this and the class below was left uncommented. 
# If the one below is buggy / wrong, try this one instead
# class RecordSignal(diacore.SignalDiagnostic):
#     def __init__(self, 
#         sig_name, 
#         *args,
#         signal_idx = None,
#         num_channels = None,
#         **kwargs):
#         super().__init__(*args, **kwargs)
#         self.sig_name = sig_name
#         self.signal_idx = signal_idx
#         self.num_channels = num_channels

#         if self.num_channels is None:
#             self.signal = np.full((self.sim_info.tot_samples), np.nan)
#         else:
#             self.signal = np.full((self.num_channels, self.sim_info.tot_samples), np.nan)
        
#     def save(self, processor, chunkInterval, globInterval):
#         if self.num_channels is not None:#processor.sig[self.sig_name].shape[0] > 1 and self.signal_idx is None:
#             self.signal[:, globInterval[0]:globInterval[1]] = \
#                 processor.sig[self.sig_name][:,chunkInterval[0]:chunkInterval[1]]
#         elif self.signal_idx is None:
#             self.signal[globInterval[0]:globInterval[1]] = \
#                 processor.sig[self.sig_name][0,chunkInterval[0]:chunkInterval[1]]
#         else:
#             self.signal[globInterval[0]:globInterval[1]] = \
#                 processor.sig[self.sig_name][self.signal_idx,chunkInterval[0]:chunkInterval[1]]

#     def get_output(self):
#         return self.signal


class RecordSignal(diacore.SignalDiagnostic):
    def __init__(self, 
        sig_name,
        *args,
        num_channels = 1, 
        channel_idx = None,
        **kwargs):
        super().__init__(*args, **kwargs)
        self.sig_name = sig_name
        self.num_channels = num_channels
        self.channel_idx = channel_idx

        if self.channel_idx is not None:
            if isinstance(self.channel_idx, int):
                self.channel_idx = (self.channel_idx,)
            assert len(self.channel_idx) == self.num_channels
        self.signal = np.full((self.num_channels, self.sim_info.tot_samples), np.nan)
        
    def save(self, processor, sig, chunkInterval, globInterval):
        assert sig[self.sig_name].ndim == 2
        #assert processor.sig[self.sig_name].shape[0] == self.num_channels
        #if processor.sig[self.sig_name].shape[0] > 1:
        #    raise NotImplementedError
        if self.channel_idx is not None:
            self.signal[:, globInterval[0]:globInterval[1]] = \
                sig[self.sig_name][self.channel_idx,chunkInterval[0]:chunkInterval[1]]
        else:
            self.signal[:, globInterval[0]:globInterval[1]] = \
                sig[self.sig_name][:,chunkInterval[0]:chunkInterval[1]]

    def get_output(self):
        return self.signal


class RecordState(diacore.StateDiagnostic):
    """
    If state_dim is 1, the state is a scalar calue
    If the number is higher, the state is a vector, and the value of each component
        will be plotted as each line by the default plot function
    
    If state_dim is a tuple (i.e. the state is a matrix/tensor), 
    then the default plot function will have trouble
    """
    def __init__(self, 
        state_name,
        state_dim,
        *args,
        label_suffix_channel = None,
        **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(state_dim, int):
            state_dim = (state_dim,)

        self.state_values = np.full((*state_dim, self.save_at.num_values), np.nan)
        self.time_indices = np.full((self.save_at.num_values), np.nan, dtype=int)

        self.get_prop = diacore.attritemgetter(state_name)
        self.diag_idx = 0

        if label_suffix_channel is not None:
            self.plot_data["label_suffix_channel"] = label_suffix_channel
        

    def save(self, processor, sig, chunkInterval, globInterval):
        assert globInterval[1] - globInterval[0] == 1
        self.state_values[:, self.diag_idx] = self.get_prop(processor)
        self.time_indices[self.diag_idx] = globInterval[0]

        self.diag_idx += 1

    def get_output(self):
        return self.state_values




class SignalPower(diacore.SignalDiagnostic):
    def __init__(self, sig_name,
                        sim_info, 
                        export_at=None,
                        sig_channels =slice(None),
                        **kwargs):
        super().__init__(sim_info, export_at, **kwargs)
        self.sig_name = sig_name
        self.sig_channels = sig_channels
        self.power = np.full((sim_info.tot_samples), np.nan)
    
    def save(self, processor, sig, chunkInterval, globInterval):
        self.power[globInterval[0]:globInterval[1]] = np.mean(np.abs(
            sig[self.sig_name][self.sig_channels,chunkInterval[0]:chunkInterval[1]])**2, axis=0)

    def get_output(self):
        return self.power


class SignalPowerRatio(diacore.SignalDiagnostic):
    def __init__(self, 
        numerator_name,
        denom_name,
        sim_info, 
        export_at=None,
        numerator_channels=slice(None),
        denom_channels = slice(None),
        **kwargs
        ):
        super().__init__(sim_info, export_at, **kwargs)
        self.numerator_name = numerator_name
        self.denom_name = denom_name
        self.power_ratio = np.full((sim_info.tot_samples), np.nan)

        self.numerator_channels = numerator_channels
        self.denom_channels = denom_channels
        
    def save(self, processor, sig, chunkInterval, globInterval):
        smoother_num = fc.create_filter(ir=np.ones((1,1,self.sim_info.output_smoothing)) / self.sim_info.output_smoothing)
        smoother_denom = fc.create_filter(ir=np.ones((1,1,self.sim_info.output_smoothing)) / self.sim_info.output_smoothing)

        num = smoother_num.process(np.mean(np.abs(sig[self.numerator_name][self.numerator_channels, chunkInterval[0]:chunkInterval[1]])**2,axis=0, keepdims=True))
        denom = smoother_denom.process(np.mean(np.abs(sig[self.denom_name][self.denom_channels, chunkInterval[0]:chunkInterval[1]])**2,axis=0, keepdims=True))

        self.power_ratio[globInterval[0]:globInterval[1]] = num / denom

    def get_output(self):
        return self.power_ratio



class StatePower(diacore.StateDiagnostic):
    def __init__(self, prop_name, 
                        sim_info, 
                        export_at=None,
                        save_frequency=None, 
                        **kwargs):
        super().__init__(sim_info, save_frequency, export_at, **kwargs)
        
        self.power = np.full((self.save_at.num_values), np.nan)
        self.time_indices = np.full((self.save_at.num_values), np.nan, dtype=int)

        self.get_prop = diacore.attritemgetter(prop_name)
        self.diag_idx = 0
        

    def save(self, processor, sig, chunkInterval, globInterval):
        assert globInterval[1] - globInterval[0] == 1
        prop_val = self.get_prop(processor)
        self.power[self.diag_idx] = np.mean(np.abs(prop_val)**2)
        self.time_indices[self.diag_idx] = globInterval[0]

        self.diag_idx += 1

    def get_output(self):
        return self.power


class StateComparison(diacore.StateDiagnostic):
    """
        compare_func should take the two states as argument, and 
        return a single value representing the comparison (distance or MSE for example)
    """
    def __init__(self, compare_func, name_state1, name_state2, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.compare_func = compare_func
        self.get_state1 = diacore.attritemgetter(name_state1)
        self.get_state2 = diacore.attritemgetter(name_state2)

        self.compare_value = np.full((self.save_at.num_values), np.nan)
        self.time_indices = np.full((self.save_at.num_values), np.nan, dtype=int)
        self.diag_idx = 0

        self.plot_data["title"] = compare_func.__name__

    def save(self, processor, sig, chunkInterval, globInterval):
        assert globInterval[1] - globInterval[0] == 1
        state1 = self.get_state1(processor)
        state2 = self.get_state2(processor)

        self.compare_value[self.diag_idx] = self.compare_func(state1, state2)
        self.time_indices[self.diag_idx] = globInterval[0]
        self.diag_idx += 1
        
    def get_output(self):
        return self.compare_value



class StateSummary(diacore.StateDiagnostic):
    """
        summary_func should take the state as argument and return a single scalar
        representing the state (norm or power for example)
    """
    def __init__(self, summary_func, state_name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.summary_func = summary_func
        self.get_state = diacore.attritemgetter(state_name)

        self.summary_value = np.full((self.save_at.num_values), np.nan)
        self.time_indices = np.full((self.save_at.num_values), np.nan, dtype=int)
        self.diag_idx = 0

        self.plot_data["title"] = summary_func.__name__
        

    def save(self, processor, sig, chunkInterval, globInterval):
        assert globInterval[1] - globInterval[0] == 1
        state = self.get_state(processor)

        self.summary_value[self.diag_idx] = self.summary_func(state)
        self.time_indices[self.diag_idx] = globInterval[0]
        self.diag_idx += 1

    def get_output(self):
        return self.summary_value





def power_of_all_signals(processor):
    """Must be called from processor.prepare(), not
        processor.__init__()
    """
    for sig_name in processor.sig.keys():
        processor.diag.add_diagnostic(f"power_{sig_name}", 
                SignalPower(sig_name, processor.sim_info, processor.block_size, 
                preprocess=[[pp.smooth(processor.sim_info.output_smoothing), pp.db_power]]))





class SoundfieldPower(diacore.Diagnostic):
    export_functions = {
        "image" : dplot.soundfield, 
        "npz" : dplot.savenpz, 
    }
    def __init__(
        self, 
        sig_name, 
        pos_mic, 
        use_samples, 
        sim_info, 
        plot_arrays = None, 
        export_func = "image", 
        export_kwargs = None, 
        preprocess = None, 
        ):
        """
        
        use_samples : tuple[int, int]
            Uses the samples from use_samples[0] (inclusive) to use_samples[1] (exclusive)
            to compute the average soundfield power. 
        
        """
        save_at = diacore.IntervalCounter((use_samples,))
        export_at = (use_samples[1],)

        #save_at_idx = [exp_idx - num_avg]
        keep_only_last_export = False
        super().__init__(sim_info, export_at, save_at, export_func, keep_only_last_export, export_kwargs, preprocess)
        self.sig_name = sig_name
        self.pos_mic = pos_mic
        self.num_mic = self.pos_mic.shape[0]
        self.plot_arrays = plot_arrays

        self.num_avg = use_samples[1] - use_samples[0]
        assert self.num_avg >= 1
        

        self.power = np.zeros(self.num_mic)

        #src_sig = {src_name : np.zeros((arrays[src_name].num)) for src_name in source_names}

    def save(self, processor, sig, chunkInterval, globInterval):
        self.power[:] += np.sum(np.abs(processor.sig[self.sig_name][:,chunkInterval[0]:chunkInterval[1]])**2, axis=-1) / self.num_avg

    def get_output(self):
        return self.power
        



import datetime




def get_time_string(detailed=False):
    tm = datetime.datetime.now()
    time_str = (
        str(tm.year)
        + "_"
        + str(tm.month).zfill(2)
        + "_"
        + str(tm.day).zfill(2)
        + "_"
        + str(tm.hour).zfill(2)
        + "_"
        + str(tm.minute).zfill(2)
    )  # + "_"+\
    # str(tm.second).zfill(2)
    if detailed:
        time_str += "_" + str(tm.second).zfill(2)
        time_str += "_" + str(tm.microsecond).zfill(2)
    return time_str


def get_unique_folder_name(prefix, parent_folder, detailed_naming=False):
    file_name = prefix + get_time_string(detailed=detailed_naming)
    file_name += "_0"
    folder_name = parent_folder.joinpath(file_name)
    if folder_name.exists():
        idx = 1
        folder_name_len = len(folder_name.name) - 2
        while folder_name.exists():
            new_name = folder_name.name[:folder_name_len] + "_" + str(idx)
            folder_name = folder_name.parent.joinpath(new_name)
            idx += 1
    # folderName += "/"
    return folder_name


def get_multiple_unique_folder_names(prefix, parent_folder, num_names):
    start_path = get_unique_folder_name(prefix, parent_folder)
    sub_folder_name = start_path.parts[-1]
    base_folder = start_path.parent

    start_idx = int(sub_folder_name.split("_")[-1])
    start_idx_len = len(sub_folder_name.split("_")[-1])
    base_name = sub_folder_name[:-start_idx_len]

    folder_names = []
    for i in range(num_names):
        folder_names.append(base_folder.joinpath(base_name + str(i + start_idx)))

    return folder_names


def get_highest_numbered_file(folder, prefix, suffix):
    highest_file_idx = -1
    for file_path in folder.iterdir():
        if file_path.name.startswith(prefix) and file_path.name.endswith(suffix):
            summary_idx = file_path.name[len(prefix) : len(file_path.name) - len(suffix)]
            try:
                summary_idx = int(summary_idx)
                if summary_idx > highest_file_idx:
                    highest_file_idx = summary_idx
            except ValueError:
                print("Warning: check prefix and suffix")

    if highest_file_idx == -1:
        return None
    else:
        fname = prefix + str(highest_file_idx) + suffix
        return folder.joinpath(fname)


def find_all_earlier_files(
    folder, name, current_idx, name_includes_idx=True, error_if_future_files_exist=True
):
    if name_includes_idx:
        name = name[: -len(str(current_idx))]
    else:
        name = name + "_"

    earlier_files = []
    for f in folder.iterdir():
        if f.stem.startswith(name) and f.stem[len(name):].isdigit():
            f_idx = int(f.stem[len(name) :])
            if f_idx > current_idx:
                if error_if_future_files_exist:
                    raise ValueError
                else:
                    continue
            elif f_idx == current_idx:
                continue
            earlier_files.append(f)
    return earlier_files


def find_index_in_name(name):
    idx = []
    for ch in reversed(name):
        if ch.isdigit():
            idx.append(ch)
        else:
            break
    if len(idx) == 0:
        return None
    idx = int("".join(idx[::-1]))
    assert name.endswith(str(idx))
    return idx


import numpy as np
import dill
import json
from abc import ABC
import copy
import matplotlib.pyplot as plt
import matplotlib.patches as patches

import aspsim.diagnostics.plot as dplot
import aspsim.room.roomimpulseresponse as rir
import aspsim.room.region as reg
import aspsim.room.trajectory as tr

class ArrayCollection():
    """ A class for managing arrays of sources and microphones, as well as the paths between them

    Attributes
    ----------
    arrays : dict of Array objects
        The arrays that have been added to the collection
    names_mic : list of str
        The names of all microphone arrays
    names_src : list of str
        The names of all source arrays
    names_free_src : list of str
        The names of the free source arrays
    paths : two-layer dict conatining ndarrays
        each path can be accessed as paths[src_name][mic_name], which
        is an ndarray of shape (num_src, num_mic, num_samples), representing
        the impulse response between a source and a microphone array
    path_type : two-layer dict containing str
        each path type can be accessed as path_type[src_name][mic_name], which
        is a string representing the type of propagation between a source and a microphone array
        See create_path for more info on possible values
    path_info : dict of dicts
        Contains metadata about the paths, such as truncation error and reverberation time.
        Can be accessed as path_info[f'{src_name}->{mic_name}']
    """
    def __init__(self):
        """Initializes an empty ArrayCollection
        """
        self.arrays = {}
        self.names_mic = []
        self.names_src = []
        self.names_free_src = []
        self.paths = {}
        self.path_type = {}

        self.path_info = {}

        self._rir_dynamic_all = [] #used to save dynamic RIRs. Temporary and for debugging only

    def __getitem__(self, key):
        return self.arrays[key]

    def __iter__(self):
        for ar in self.arrays.values():
            yield ar
    
    def __contains__(self, arrayName):
        return arrayName in self.arrays

    def save_metadata(self, folder_path):
        """Saves all metadata to the given folder
        
        Parameters
        ----------
        filepath : pathlib.Path
        """
        if folder_path is not None:
            self._save_metadata_arrays(folder_path)
            self._save_metadata_paths(folder_path)
            self._save_readable_pos(folder_path)
        
    def _save_metadata_arrays(self, filepath):
        array_info = {}
        for ar_name, ar in self.arrays.items():
            array_info[ar_name] = ar.metadata
        with open(filepath.joinpath("metadata_arrays.json"), "w") as f:
            json.dump(array_info, f, indent=4)

    def _save_metadata_paths(self, filepath):
        #path_info = {}
        #for src, mic in self.mic_src_combos():
        #    path_info[f"{src.name}->{mic.name}"] = self.path_type[src.name][mic.name]
        with open(filepath.joinpath("metadata_paths.json"), "w") as f:
            json.dump(self.path_info, f, indent=4)

    def _save_readable_pos(self, filepath):
        pos = {}
        for ar_name, ar in self.arrays.items():
            pos[ar_name] = ar.pos.tolist()
        with open(filepath.joinpath("array_pos.json"), "w") as f:
            json.dump(pos, f, indent=4)


    def set_default_path_type(self, path_type):
        """Sets the path type for all paths that were not modified by the user

        Parameters
        ----------
        path_type : str
            The path type to be set. See create_path for more info on possible values
        """
        for src, mic in self.mic_src_combos():
            if not mic.name in self.path_type[src.name] and not mic.name in self.paths[src.name]:
                self.path_type[src.name][mic.name] = path_type
            elif mic.name in self.paths[src.name]:
                self.path_type[src.name][mic.name] = "modified"

    def empty(self):
        """Returns True if the collection has no arrays
        """
        return len(self.arrays) == 0

    def sources(self):
        """Iterates over all source arrays"""
        for name in self.names_src:
            yield self.arrays[name]
    
    def mics(self):
        """Iterates over all mic arrays"""
        for name in self.names_mic:
            yield self.arrays[name]

    def free_sources(self):
        """Iterates over all free source arrays"""
        for name in self.names_free_src:
            yield self.arrays[name]

    def mic_src_combos(self):
        """Iterates over the all combinations of mics and sources"""
        for src_name in self.names_src:
            for mic_name in self.names_mic:
                yield self.arrays[src_name], self.arrays[mic_name]

    def iter_paths(self):
        """Iterates over all paths"""
        for src_name in self.names_src:
            for mic_name in self.names_mic:
                yield self.arrays[src_name], self.arrays[mic_name], self.paths[src_name][mic_name]

    def add_array(self, array):
        """Adds an array to the collection
        
        Parameters
        ----------
        array : Array
            The array to be added. Can be any subclass of Array
        """
        assert array.name not in self.arrays
        self.arrays[array.name] = array
        
        if array.is_mic:
            self.names_mic.append(array.name)
        elif array.is_source:
            self.names_src.append(array.name)
            self.paths[array.name] = {}
            self.path_type[array.name] = {}
            if isinstance(array, FreeSourceArray):
                self.names_free_src.append(array.name)
        else:
            raise ValueError("Array is neither source nor microphone")

    def set_prop_paths(self, paths):
        """Sets the propagation paths between some sources and microphones

        Parameters
        ----------
        paths : dict of dicts of ndarrays
            paths[src_name][mic_name] is an ndarray of shape (num_src, num_mic, num_samples)
            representing the impulse response between a source and a microphone array
            the source and and mic array must already be in the collection
        """
        for src_name, src_paths in paths.items():
            for mic_name, path in src_paths.items():
                self.set_prop_path(path, src_name, mic_name)

    def set_prop_path(self, path, src_name, mic_name):
        """Sets the propagation path between a source and a microphone array
        
        Parameters
        ----------
        path : ndarray of shape (num_src, num_mic, num_samples)
            representing the impulse response between a source and a microphone array
            The ndarray must be of the correct shape
        src_name : str
            name of the source array. 
            The array must already be in the collection
        mic_name : str
            name of the microphone array. 
            The array must already be in the collection
        """
        assert src_name in self.names_src
        assert mic_name in self.names_mic
        assert isinstance(path, np.ndarray)
        assert path.shape[0] == self.arrays[src_name].num
        assert path.shape[1] == self.arrays[mic_name].num
        assert path.ndim == 3
        self.paths[src_name][mic_name] = path
    
    def set_path_types(self, path_types):
        """Sets the propagation type between some sources and microphones
        
        Parameters
        ----------
        path_types : dict of dicts of str
            path_types[src_name][mic_name] is a string representing the type of propagation
            between a source and a microphone array
            The source and and mic array must already be in the collection
            See create_path for more info on possible values
        """
        for src_name, src_path_types in path_types.items():
            for mic_name, pt in src_path_types.items():
                assert src_name in self.names_src
                assert mic_name in self.names_mic
                assert isinstance(pt, str)
                self.path_type[src_name][mic_name] = pt


    
# def setup_ir(self, sim_info):
#     """Generates the impulse responses between all sources and microphones

#     The method uses the path_type attribute to determine the type of propagation
    
#     Parameters
#     ----------
#     sim_info : SimInfo
#         The simulation info object
#     """
#     rir.setup_ir(self, sim_info)

    def setup_ir(self, sim_info):
        """Generates the impulse responses between all sources and microphones

        The method uses the path_type attribute to determine the type of propagation
        
        Parameters
        ----------
        sim_info : SimInfo
            The simulation info object
        """
        metadata = {}
        self.sim_info = sim_info
        self.path_generator = rir.PathGenerator(self.sim_info, self)

        for src, mic in self.mic_src_combos():
            self.path_info[f"{src.name}->{mic.name}"] = {}

            if src.num == 0 or mic.num == 0:
                reverb = "none"
            else:
                reverb = self.path_type[src.name][mic.name]

            self.path_info[f"{src.name}->{mic.name}"]["type"] = reverb
            print(f"{src.name}->{mic.name} has propagation type: {reverb}")
            if reverb != "modified":
                self.paths[src.name][mic.name], path_info = self.path_generator.create_path(src, mic, reverb, sim_info, True, True)
                for key, val in path_info.items():
                    self.path_info[f"{src.name}->{mic.name}"][key] = val
        

    def update_path(self, src, mic):
        """Updates the path between a source and a microphone array
        
        Parameters
        ----------
        src : Array
            The source array
        mic : Array
            The microphone array
        """
        reverb = self.path_type[src.name][mic.name]
        assert reverb != "modified"
        self.paths[src.name][mic.name] = self.path_generator.create_path(src, mic, reverb, self.sim_info)

    def update(self, glob_idx):
        """Updates the arrays and paths that change over time

        The method performs the following steps
        1. update arrays pos/properties
        2. get info about which arrays actually changed
        3. update the paths connecting the updated arrays
        
        Parameters
        ----------
        glob_idx : int
            The global time index of the current sample
        """
        
        if glob_idx % self.sim_info.array_update_freq != 0:
            return 

        changed_arrays = []

        for ar_name, ar in self.arrays.items():
            changed = ar.update(glob_idx)
            if changed:
                changed_arrays.append(ar_name)

        already_updated = []
        
        for ar_name in changed_arrays:
            if self.arrays[ar_name].is_mic:
                for src in self.sources():
                    if not src.name in already_updated:
                        self.update_path(src, self.arrays[ar_name])
                self._rir_dynamic_all.append(self.paths[src.name][ar_name]) # for debugging only
            elif self.arrays[ar_name].is_source:
                for mic in self.mics():
                    if not mic.name in already_updated:
                        self.update_path(self.arrays[ar_name], mic)
            else:
                raise ValueError("Array must be mic or source")
            
            already_updated.append(ar_name)



    def plot(self, sim_info, fig_folder, print_method):
        """Creates a plot with all arrays in the collection
        
        If ISM is chosen as the propagation type, the room is also plotted
        
        Parameters
        ----------
        sim_info : SimInfo
            The simulation info object
        fig_folder : pathlib.Path
            The folder where the figure should be saved
        print_method : str
            The method used to save the figure. See aspsim.diagnostics.plot.output_plot
            for a list of the options
        """
        fig, ax = plt.subplots()
        for ar in self.arrays.values():
            ar.plot(ax, sim_info)

        if "ism" in [self.path_type[src.name][mic.name] for src, mic in self.mic_src_combos()]:
            corner = [c - sz/2 for c, sz in zip(sim_info.room_center[:2], sim_info.room_size[:2])]
            #bottom_left_corner = sim_info.room_center[:2] - (sim_info.room_size[:2] / 2)
            width = sim_info.room_size[0]
            height = sim_info.room_size[1]
            ax.add_patch(patches.Rectangle(corner, width, height, edgecolor="k", facecolor="none", linewidth=2, alpha=1))

        ax.legend()
        ax.axis("equal")
        ax.grid(True)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        dplot.output_plot(print_method, fig_folder, "array_pos")

    def save_to_file(self, folder_path):
        """Saves the array collection to file
        
        Parameters
        ----------
        folder_path : pathlib.Path
            The path to the folder where the collection should be saved
        """
        if folder_path is not None:
            with open(folder_path.joinpath("arrays.pickle"), "wb") as f:
                dill.dump(self, f)

    def get_freq_paths(self, num_freqs, samplerate):
        """Get the frequency domain response of the paths between all sources and microphones
        
        Parameters
        ----------
        arrays : ArrayCollection
            The array collection to get the frequency paths from
        num_freqs : int
            The number of frequencies to calculate the response for. The number refers to the total FFT
            length, so the number of real frequencies is num_freqs // 2 + 1, which is the number of 
            frequency bins in the output. 
        
        Returns
        -------
        freq_paths : dict of dicts of ndarrays
            freq_paths[src_name][mic_name] is a complex ndarray of shape (num_real_freqs, num_mic, num_src)
            representing the frequency response between a source and a microphone array
        freqs : ndarray of shape (num_real_freqs,)
            The real frequencies in Hz of the response

        Notes
        -----
        The frequencies returned are compatible with get_real_freqs of the package aspcol, as well as the
        output of np.fft.rfft. 

        num_freqs can be safely chosen as a larger number than the number of samples in the impulse response,
        as the FFT will zero-pad the signal to the desired length. But if num_freqs is smaller, then the
        signal will be truncated.
        """
        if num_freqs % 2 == 1:
            raise NotImplementedError("Odd number of frequencies not supported yet")
        
        freqs = (samplerate / (num_freqs)) * np.arange(num_freqs // 2 + 1)

        fpaths = {}
        for src, mic, path in self.iter_paths():
            fpaths.setdefault(src.name, {})
            fpaths[src.name][mic.name] = np.fft.rfft(path, n=num_freqs, axis=-1).T
                
        return fpaths, freqs

def load_arrays(folder_path):
    """Loads the array collection from file
    
    Parameters
    ----------
    folder_path : pathlib.Path
        The path to the folder where the collection should be loaded from
        The file is assumed to be named arrays.pickle
    """
    with open(folder_path.joinpath("arrays.pickle"), "rb") as f:
        arrays = dill.load(f)
    return arrays


def prototype_equals(prototype, initialized):
    """Compares an initialized collection to a prototype collection.

    This function checks if the initialized collection could be created from the prototype. i.e. if it 
        would make sense to load initialized instead of generating the paths of the prototype

    Parameters
    ----------
    prototype : ArrayCollection
        Trototype collection need not have the paths initialized, but must 
        have arrays added with positions set. 
    initialized : ArrayCollection
        The initialized collection.
        """
    #check that they have the same array names
    if prototype.arrays.keys() != initialized.arrays.keys():
        return False

    #Check that the arrays are the same
    for ar_name, ar in prototype.arrays.items():
        #if ar != initialized[ar_name]:
        #    return False
        if type(ar) is not type(initialized[ar_name]):
            return False
        if ar.num != initialized[ar_name].num:
            return False
        if not np.allclose(ar.pos,initialized[ar_name].pos):
            return False

    #Check the paths are the same
    for src, mic in prototype.mic_src_combos():
        if prototype.path_type[src.name][mic.name] == "modified":
            if not np.allclose(prototype.paths[src.name][mic.name], initialized.paths[src.name][mic.name]):
                return False
        if prototype.path_type[src.name][mic.name] != initialized.path_type[src.name][mic.name]:
            return False

    return True

class Array(ABC):
    is_mic = False
    is_source = False
    plot_symbol = "."
    def __init__(self, name, pos, directivity_type=None, directivity_dir=None):
        """Abstract base class for all types of arrays. Establishes
        the common attributes and methods of all array types

        Parameters
        ----------
        name : str
            name of the array, which will also be the name of the
            signal associated with the array
        pos : ndarray of shape (num_objects, spatial_dim)
            or a list of ndarrays of the same shape, which will distribute
                the objects into groups.
            or a Trajectory object, which will produce a moving array
        directivity_type : optional, list of str
            each string is one of "omni", "cardioid", "hypercardioid", 
            "supercardioid", "bidirectional". 
            If supplied, directivity_dir must also be supplied
            if not supplied, the array will be omnidirectional
        directivity_dir : optional, ndarray of shape (num_objects, spatial_dim)
            each row is a unit vector pointing in the direction of the directivity for that object
            If supplied, directivity_type must also be supplied
        Notes
        -----
        Documentation refers to objects, which means any of microphone, loudspeaker
        or source depending on array type. 
        """
        self.name = name

        if isinstance(pos, (list, tuple)):
            assert all([isinstance(p, np.ndarray) for p in pos])
            num_in_group = np.cumsum([0] + [p.shape[0] for p in pos])
            self.group_idxs = [np.arange(num_in_group[i], num_in_group[i+1]) for i in range(len(num_in_group)-1)]
            self.num_groups = len(pos)
            self.pos = np.concatenate(pos, axis=0)
            self.pos_segments = pos
            self.dynamic = False
        elif isinstance(pos, np.ndarray):
            self.group_idxs = None
            self.num_groups = 1
            self.pos = pos
            self.pos_segments = [pos]
            self.dynamic = False
        elif isinstance(pos, tr.Trajectory):
            self.trajectory = pos
            self.group_idxs = None
            self.num_groups = 1
            self.pos = self.trajectory.current_pos(0)
            self.pos_segments = [pos]
            self.dynamic = True
            self.pos_all = []
            self.time_all = []
        else:
            raise ValueError("Incorrect datatype for pos")

        self.num = self.pos.shape[0]
        assert self.pos.ndim == 2

        if directivity_type is None or directivity_dir is None:
            assert directivity_type is None and directivity_dir is None
            directivity_type = ["omni" for _ in range(self.num)]
        else:
            assert directivity_type is not None and directivity_dir is not None
            assert len(directivity_type) == self.num
            assert directivity_dir.shape[0] == self.num
            assert all([d in ["omni", "cardioid", "hypercardioid", "supercardioid", "bidirectional"] for d in directivity_type])
            assert np.allclose(np.linalg.norm(directivity_dir, axis=-1), 1)
        self.directivity_type = directivity_type
        self.directivity_dir = directivity_dir

        self.metadata = {
            "type" : self.__class__.__name__,
            "number" : self.num,
            "number of groups" : self.num_groups,
            "dynamic" : self.dynamic,
        }

    def set_groups(self, group_idxs):
        """Sorts each object of the array into groups

        Parameters
        ----------
        group_idxs : list of lists or list of 1D nd.arrays
            Each inner list is the indices of the array elements
            belonging to one group.
        """
        self.num_groups = len(group_idxs)
        self.group_idxs = group_idxs

    def plot(self, ax, sim_info):
        if self.dynamic:
            self.trajectory.plot(ax, self.plot_symbol, self.name, sim_info.tot_samples)
        else:
            ax.plot(self.pos[:,0], self.pos[:,1], self.plot_symbol, label=self.name, alpha=0.8)

    def update(self, glob_idx):
        if self.dynamic:
            self.pos = self.trajectory.current_pos(glob_idx)
            self.time_all.append(glob_idx)
            self.pos_all.append(self.pos)
            return True
        return False

class MicArray(Array):
    """Array class for microphones
    
    Parameters
    ----------
    name : str
        name of the array, which will also be the name of the
        signal associated with the array
    pos : ndarray of shape (num_mics, spatial_dim)
        position of the microphones
    """
    is_mic = True
    plot_symbol = "x"
    def __init__(self, name, pos, **kwargs):
        super().__init__(name, pos, **kwargs)

class RegionArray(MicArray):
    """Class for representing a continuous region with an microphone array
    
    Requires a Region object to be used as representation of the region. The Region object
    has a equally_spaced_points method, which is used to generate the positions of the microphones
    unless the pos argument is supplied. In that case, the signals generated will be 
    identical to using a MicArray() with region.equally_spaced_points() as positions. 

    Attributes
    ----------
    region : Region
        The region object used to represent the region
    region_segments : list of Region
        If the region is a CombinedRegion, this attribute will contain the individual regions
        that were combined to create the region. 
    pos : ndarray of shape (num_mics, spatial_dim)
        The positions of the microphones that are used in the simulation
    """
    def __init__(self, name, region, pos=None, **kwargs):
        if isinstance(region, (list, tuple)):
            if len(region) > 1:
                self.region_segments = region
                region = reg.CombinedRegion(region)
            else:
                self.region_segments = region
                region = region[0]
        else:
            self.region_segments = [region]

        if pos is None:
            pos = [r.equally_spaced_points() for r in self.region_segments]
        super().__init__(name, pos, **kwargs)
        self.region = region
        self.metadata["region shape"] = self.region.__class__.__name__

    def plot(self, ax, sim_info):
        self.region.plot(ax, self.name)
        
class ControllableSourceArray(Array):
    """ Array for sources controllable by a processor
    
    Parameters
    ----------
    name : str
        name of the array, which will also be the name of the
        signal associated with the array
    pos : ndarray of shape (num_sources, spatial_dim)
        position of the sources
    
    """
    is_source = True
    plot_symbol="o"
    def __init__(self, name, pos, **kwargs):
        super().__init__(name, pos, **kwargs)
        
class FreeSourceArray(Array):
    """ Array for free sound sources, that cannot be adaptively controlled 
    
    Parameters
    ----------
    name : str
        name of the array, which will also be the name of the
        signal associated with the array
    pos : ndarray of shape (num_sources, spatial_dim)
        position of the sources
    source : Source
        The source object used to generate the signal
        see aspsim.signal.source module for more info. More sources are also available
        in the aspsim.room.sourcescollection module. 

    """
    is_source = True
    plot_symbol = "s"
    def __init__(self, name, pos, source, **kwargs):
        super().__init__(name, pos, **kwargs)
        self.set_source(source)

        self.metadata["source info"] = self.source.metadata
    
    def reset_state(self):
        self.source = copy.deepcopy(self.source)
        self.source.reset()

    def get_samples(self, num_samples):
        return self.source.get_samples(num_samples)

    def set_source(self, source):
        assert source.num_channels == self.num
        self.source = source






import numpy as np
import scipy.signal as spsig

import aspcore.filterdesign as fd



def differential_cardioid_microphone(signal_main, signal_secondary, microphone_distance, filter_order, c, samplerate, filter_below=None):
    """Constructs a cardioid microphone response from two omnidirectional microphones
    
    The two omnidirectional microphones must be separated by a distance of microphone_distance.
    The strongest response is in the direction of the main microphone from the secondary, and the null is in the opposite direction.

    Parameters
    ----------
    signal_main : ndarray of shape (num_mics, num_samples)
        the signal from the main microphone
    signal_secondary : ndarray of shape (num_mics, num_samples)
        the signal from the secondary microphone
    microphone_distance : float
        the distance between the microphones
    filter_order : int
        the order of the filter. The filter will have 2*filter_order + 1 taps. 
    c : float
        the speed of sound. 
    samplerate : int
        the sampling rate of the signals. 
    filter_below : float or None
        if not None, filters out the frequencies below this frequency, given in Hertz. The differential microphone
        is problematic at low frequencies, and this can be used to avoid this.

    Returns
    -------
    cardioid_signal : ndarray of shape (num_mics, num_samples)
        the signal from the cardioid microphone

    References
    ----------
    [benestyStudy2013] J. Benesty and J. Chen, Study and design of differential microphone arrays, vol. 6. in Springer Topics in Signal Processing, vol. 6. Springer, 2013.
    """
    num_freq = 4*filter_order # we need more or equal frequency samples compared to the ir_len
    ir_len = 2*filter_order + 1
    #wave_num = ft.get_real_wavenum(num_freq, samplerate, c)

    def freq_response_main_mic(f):
        wave_num = 2 * np.pi * f / c
        wave_num[wave_num == 0] = 1e-10 # To avoid division by zero warnings. Not strictly necessary.
        both_mic_factor = 1 / (1 - np.exp(-1j * 2 * wave_num * microphone_distance))
        both_mic_factor[0] = 0
        mic1_factor = np.conj(both_mic_factor)

        if filter_below is not None:
            num_freqs_to_filter = np.sum(f <= filter_below)
            scaling = np.linspace(0, 1, num_freqs_to_filter) ** 4
            mic1_factor[:num_freqs_to_filter] *= scaling
        return mic1_factor
    
    def freq_response_secondary_mic(f):
        wave_num = 2 * np.pi * f / c
        wave_num[wave_num == 0] = 1e-10 # To avoid division by zero warnings. Not strictly necessary.

        both_mic_factor = 1 / (1 - np.exp(-1j * 2 * wave_num * microphone_distance))
        both_mic_factor[0] = 0
        mic2_factor = -np.conj(both_mic_factor * np.exp(-1j * wave_num * microphone_distance))

        if filter_below is not None:
            num_freqs_to_filter = np.sum(f <= filter_below)
            scaling = np.linspace(0, 1, num_freqs_to_filter) ** 4
            mic2_factor[:num_freqs_to_filter] *= scaling
        return mic2_factor
    
    mic1_ir = fd.fir_from_frequency_function(freq_response_main_mic, ir_len, samplerate, window=None)
    mic2_ir = fd.fir_from_frequency_function(freq_response_secondary_mic, ir_len, samplerate, window=None)

    #mic1_factor = ft.insert_negative_frequencies(mic1_factor, even = True)
    #mic2_factor = ft.insert_negative_frequencies(mic2_factor, even = True)

    #mic1_ir, _ = fir_from_freqs_window(mic1_factor, ir_len, two_sided=True, window="hamming")
    #mic2_ir, _ = fir_from_freqs_window(mic2_factor, ir_len, two_sided=True, window="hamming")

    signal_main_filtered = spsig.fftconvolve(mic1_ir[None,:], signal_main, axes=-1, mode="full")
    signal_secondary_filtered = spsig.fftconvolve(mic2_ir[None,:], signal_secondary, axes=-1, mode="full")

    cardioid_signal = signal_main_filtered + signal_secondary_filtered
    cardioid_signal = cardioid_signal[...,filter_order:filter_order+signal_main.shape[-1]]
    return cardioid_signal





# def differential_cardioid_microphone(signal_main, signal_secondary, microphone_distance, filter_order, c, samplerate, filter_below=None):
#     """Constructs a cardioid microphone response from two omnidirectional microphones
    
#     The two omnidirectional microphones must be separated by a distance of microphone_distance.
#     The strongest response is in the direction of the main microphone from the secondary, and the null is in the opposite direction.

#     Parameters
#     ----------
#     signal_main : ndarray of shape (num_mics, num_samples)
#         the signal from the main microphone
#     signal_secondary : ndarray of shape (num_mics, num_samples)
#         the signal from the secondary microphone
#     microphone_distance : float
#         the distance between the microphones
#     filter_order : int
#         the order of the filter. The filter will have 2*filter_order + 1 taps. 
#     c : float
#         the speed of sound. 
#     samplerate : int
#         the sampling rate of the signals. 
#     filter_below : float or None
#         if not None, filters out the frequencies below this frequency, given in Hertz. The differential microphone
#         is problematic at low frequencies, and this can be used to avoid this.

#     Returns
#     -------
#     cardioid_signal : ndarray of shape (num_mics, num_samples)
#         the signal from the cardioid microphone

#     References
#     ----------
#     [benestyStudy2013] J. Benesty and J. Chen, Study and design of differential microphone arrays, vol. 6. in Springer Topics in Signal Processing, vol. 6. Springer, 2013.
#     """
#     num_freq = 4*filter_order # we need more or equal frequency samples compared to the ir_len
#     ir_len = 2*filter_order + 1
#     wave_num = ft.get_real_wavenum(num_freq, samplerate, c)

#     #def freq_response_main_mic(f):
#     both_mic_factor = 1 / (1 - np.exp(-1j * 2 * wave_num * microphone_distance))
#     both_mic_factor[0] = 0
#     mic1_factor = np.conj(both_mic_factor)
#     mic2_factor = -np.conj(both_mic_factor * np.exp(-1j * wave_num * microphone_distance))

#     if filter_below is not None:
#         #def freq_drop_function(freq):
#         #    return 1 - 1 / (1 + (freq / filter_below) ** 4)
        
#         freqs = ft.get_real_freqs(num_freq, samplerate)
#         num_freqs_to_filter = np.sum(freqs <= filter_below)
#         #freqs_to_filter = freqs[freqs_to_filter_mask]
#         scaling = np.linspace(0, 1, num_freqs_to_filter)
#         scaling = scaling ** 4
#         mic1_factor[:num_freqs_to_filter] *= scaling
#         mic2_factor[:num_freqs_to_filter] *= scaling

#     mic1_factor = ft.insert_negative_frequencies(mic1_factor, even = True)
#     mic2_factor = ft.insert_negative_frequencies(mic2_factor, even = True)

#     mic1_ir, _ = fir_from_freqs_window(mic1_factor, ir_len, two_sided=True, window="hamming")
#     mic2_ir, _ = fir_from_freqs_window(mic2_factor, ir_len, two_sided=True, window="hamming")

#     signal_main_filtered = signal.fftconvolve(mic1_ir[None,:], signal_main, axes=-1, mode="full")
#     signal_secondary_filtered = signal.fftconvolve(mic2_ir[None,:], signal_secondary, axes=-1, mode="full")

#     cardioid_signal = signal_main_filtered + signal_secondary_filtered
#     cardioid_signal = cardioid_signal[...,filter_order:filter_order+signal_main.shape[-1]]
#     return cardioid_signal


import numpy as np
from abc import ABC, abstractmethod

import scipy.signal as spsig
import scipy.linalg as splin
import soundfile as sf

import aspsim.utilities as util

import aspcore.correlation as cr
import aspcore.utilities as coreutil
import aspcore.filter as fc

class Source(ABC):
    def __init__(self, num_channels, rng=None):
        self.num_channels = num_channels

        if rng is None:
            self.rng = np.random.default_rng(123456)
        else:
            self.rng = rng

        self.metadata = {}
        self.metadata["number of channels"] = self.num_channels
        #self.metadata["power"] = self.power
    
    @abstractmethod
    def get_samples(self, num_samples):
        return np.zeros((self.num_channels, num_samples))

class SourceArray:
    def __init__(self, source_type, num_sources, amplitude, *args):
        if isinstance(amplitude, (int, float)):
            amplitude = [amplitude for _ in range(num_sources)]

        self.sources = [source_type(amplitude[i], *args) for i in range(num_sources)]

    def get_samples(self, num_samples):
        output = np.concatenate(
            [src.get_samples(num_samples) for src in self.sources], axis=0
        )
        return output



class Sequence(Source):
    def __init__(self, audio, amp_factor = 1, end_mode = "repeat"):
        """ Will play the supplied sequence

        Parameters
        ----------
        audio is np.ndarray of audio samples to be played. 
                    Shape is (num_channels, num_samples)
        end_mode : can be any of {'repeat', 'zeros', 'raise'}
            with 'repeat' the sequence starts over again indefinitely
            with 'zeros' the sequence plays once and then only 0s 
            with 'raise' the source will raise an exception if get_samples() 
            is called after the sequence is finished
        """
        assert isinstance(audio, np.ndarray)
        if audio.ndim == 1:
            audio = audio[None,:]
        elif audio.ndim != 2:
            raise ValueError
        super().__init__(audio.shape[0])

        self.audio = audio
        self.amp_factor = amp_factor
        self.end_mode = end_mode
        self.tot_samples = audio.shape[1]
        self.current_sample = 0

    def get_samples(self, num_samples):
        sig = np.zeros((self.num_channels, num_samples))
        if self.end_mode == "repeat":
            block_lengths = util.calc_block_sizes(num_samples, self.current_sample, self.tot_samples)
            i = 0
            for block_len in block_lengths:
                sig[:,i:i+block_len] = self.audio[:,self.current_sample:self.current_sample+block_len]
                self.current_sample = (self.current_sample + block_len) % self.tot_samples
                i += block_len
        elif self.end_mode == "zeros":
            raise NotImplementedError
        elif self.end_mode == "raise":
            if self.current_sample + num_samples >= self.tot_samples:
                raise StopIteration("End of audio signal")

            sig = self.audio[:,self.current_sample:self.current_sample+num_samples]
            self.current_sample += num_samples
            return sig
        else:
            raise ValueError("Invalid end mode")
        return sig * self.amp_factor


class WhiteNoiseSource(Source):
    def __init__(self, num_channels, power, rng=None):
        super().__init__(num_channels, rng)
        self.set_power(power)

        if isinstance(self.power, (np.ndarray)):
            self.metadata["power"] = self.power.tolist()
        else:
            self.metadata["power"] = self.power

    def get_samples(self, num_samples):
        return self.rng.normal(
            loc=0, scale=self.stdDev, size=(num_samples, self.num_channels)
        ).T
    
    def set_power(self, newPower):
        self.power = newPower
        self.stdDev = np.sqrt(newPower)

        if isinstance(self.power, np.ndarray):
            assert self.power.ndim == 1
            assert self.power.shape[0] == self.num_channels or \
                    self.power.shape[0] == 1


class SineSource(Source):
    def __init__(self, num_channels, power, freq, samplerate, rng=None):
        super().__init__(num_channels, rng)
        self.power = power
        self.amplitude = np.sqrt(2 * self.power)
        # p = a^2 / 2
        # 2p = a^2
        # sqrt(2p) = a
        self.freq = freq
        self.samplerate = samplerate
        self.phase = self.rng.uniform(low=0, high=2 * np.pi, size=(self.num_channels,1))

        self.phase_per_sample = 2 * np.pi * self.freq / self.samplerate

        if isinstance(self.power, (np.ndarray)):
            self.metadata["power"] = self.power.tolist()
        else:
            self.metadata["power"] = self.power
        self.metadata["frequency"] = self.freq

    def get_samples(self, num_samples):
        noise = (
            self.amplitude
            * np.cos(self.phase_per_sample * np.arange(num_samples)[None,:] + self.phase)
        )
        self.phase = (self.phase + num_samples * self.phase_per_sample) % (2 * np.pi)
        return noise


class Counter(Source):
    """
    Counts from start_number and adds one per sample
        used primarily for debugging
    """
    def __init__(self, num_channels, start_number=0):
        super().__init__(num_channels)
        if num_channels > 1:
            raise NotImplementedError
        self.current_number = start_number

        self.metadata["start number"] = start_number 

    def get_samples(self, num_samples):
        values = np.arange(self.current_number, self.current_number+num_samples)
        self.current_number += num_samples
        return values




class MultiSineSource(Source):
    def __init__(self, num_channels, power, freq, samplerate, rng=None):
        super().__init__(num_channels, rng)
        if num_channels > 1:
            raise NotImplementedError
        if isinstance(power, (list, tuple, np.ndarray)):
            raise NotImplementedError
        if isinstance(freq, (list, tuple)):
            freq = np.array(freq)
        assert freq.ndim == 1
        self.power = power
        self.freq = freq
        self.samplerate = samplerate
        self.numSines = len(freq)
        self.phase = self.rng.uniform(low=0, high=2 * np.pi, size=self.numSines)
        self.phasePerSample = 2 * np.pi * self.freq / self.samplerate
        self.amplitude = np.sqrt(2 * self.power / self.numSines)

        if isinstance(self.power, (np.ndarray)):
            self.metadata["power"] = self.power.tolist()
        else:
            self.metadata["power"] = self.power
        self.metadata["frequency"] = self.freq

    def get_samples(self, num_samples):
        noise = np.zeros((1, num_samples))
        for i in range(self.numSines):
            noise += (
                self.amplitude
                * np.cos(
                    np.arange(num_samples) * self.phasePerSample[i] + self.phase[i]
                )[None, :]
            )
            self.phase[i] = (self.phase[i] + num_samples * self.phasePerSample[i]) % (
                2 * np.pi
            )
        return noise



class BandlimitedNoiseSource(Source):
    def __init__(self, num_channels, power, freqLim, samplerate, rng=None):
        super().__init__(num_channels, rng)
        assert len(freqLim) == 2
        assert freqLim[0] < freqLim[1]
        self.num_channels = num_channels

        wp = [freq for freq in freqLim]
        self.filtCoef = spsig.butter(16, wp, btype="bandpass", output="sos", fs=samplerate)

        testSig = spsig.sosfilt(self.filtCoef, self.rng.normal(size=(self.num_channels,10000)))
        self.testSigPow = np.mean(testSig ** 2)
        self.set_power(power)

        self.zi = spsig.sosfilt_zi(self.filtCoef)
        self.zi = np.tile(np.expand_dims(self.zi,1), (1,self.num_channels,1))

        if isinstance(self.power, (np.ndarray)):
            self.metadata["power"] = self.power.tolist()
        else:
            self.metadata["power"] = self.power
        self.metadata["frequency span"] = freqLim

    def get_samples(self, num_samples):
        noise = self.amplitude * self.rng.normal(size=(self.num_channels, num_samples))
        filtNoise, self.zi = spsig.sosfilt(self.filtCoef, noise, zi=self.zi, axis=-1)
        return filtNoise

    def set_power(self, newPower):
        self.power = newPower
        self.amplitude = np.sqrt(newPower / self.testSigPow)

        # if isinstance(self.power, np.ndarray):
        #     assert self.power.ndim == 1
        #     assert self.power.shape[0] == self.num_channels or \
        #             self.power.shape[0] == 1



class PulseTrain(Source):
    def __init__(self, num_channels, amplitude, period_len, delay=0):
        super().__init__(num_channels)

        if isinstance(amplitude, (tuple, list, np.ndarray)):
            assert len(amplitude) == self.num_channels
            self.amplitude = np.array(amplitude)
        else:
            self.amplitude = np.ones(self.num_channels) * amplitude

        if isinstance(period_len, (tuple, list, np.ndarray)):
            assert len(period_len) == self.num_channels
            self.period_len = np.array(period_len)
        else:
            self.period_len = np.ones(self.num_channels)*period_len

        if isinstance(delay, (tuple ,list, np.ndarray)):
            assert len(delay) == self.num_channels
            self.delay = np.array(delay)
        else:
            self.delay = np.ones(self.num_channels) * delay

        self.idx = np.zeros(self.num_channels)
        self.idx -= self.delay

    def get_samples(self, num_samples):
        sig = np.zeros((self.num_channels, num_samples))

        for i in range(num_samples):
            add_pulse = np.mod(self.idx, self.period_len) == 0
            sig[add_pulse,i] = self.amplitude[add_pulse]
            self.idx += 1
        return sig





def load_audio_file(file_name, desired_sr = None, num_channels=None, verbose=False):
    audio, audio_sr = sf.read(file_name)
    assert audio.ndim == 2
    audio = audio.T
    #raise NotImplementedError("Check which axis is audio and which is channel. \
    #                            Make support for multiple channels (should be simple)")
    if num_channels is not None:
        audio = audio[:num_channels, :]

    if desired_sr != audio_sr:
        assert audio_sr > desired_sr
        if audio_sr % desired_sr == 0:
            down_factor = audio_sr // desired_sr
            up_factor = 1
        else:
            up_factor, down_factor = coreutil.simplify_ratio(desired_sr, audio_sr)
        audio_downsampled = [spsig.resample_poly(audio[ch,:], up=up_factor, down=down_factor, axis=-1) for ch in range(num_channels)]
        audio = np.stack(audio_downsampled, axis=0)
        # srAudio = srAudio // downsamplingFactor
        if verbose:
            print(f"AudioFileSource downsampled audio file from \
                    {audio_sr} to {audio_sr // down_factor} Hz")

    return audio






def ar_coeffs_from_autocorr(autocorr):
    """
    returns the parameter vector a = [a_1, ..., a_p]
        that implements the AR system
        y(n) = a_1 y(n-1) + a_2 y(n-2) + ... + a_p y(n-p) + s(n)

        where s(n) is a white noise process with variance awgn_power

    """
    assert autocorr.ndim <= 2
    if autocorr.ndim == 2:
        assert autocorr.shape[0] == 1
        autocorr = autocorr[0,:]
    ar_coeffs = splin.solve_toeplitz(autocorr[:-1], autocorr[1:])
    noise_power = autocorr[0] - np.sum(autocorr[1:] * ar_coeffs)
    return ar_coeffs, noise_power

def ar_coeffs_to_transfer_function(ar_coeffs):
    """
    Gives the rational transfer function that will give rise to 
        the provided AR process, if a white noise is filtered
        through the returned filter.

    According to scipy.signal.lfilter which implements the system
        a_0 y(n) = b_0 x(n) + b_1 x(n-1) + ... + b_t x(n-t) 
                    -a_1 y(n-1) - a_2 y(n-2) - ... - a_p y(n-p)
    """

    denom_coeffs = np.concatenate(([1], -ar_coeffs))
    num_coeffs = np.array([1], dtype=float)
    return num_coeffs, denom_coeffs 

class AutocorrSource(Source):
    """

    autocorr is array of shape (num_channels, corr_len)
        with autocorr[:,0] being lag 0. 

    All channels are all independent. 
    Could be extended in the future to accept autocorr of shape 
        (num_channels, num_channels, corr_len) which could 
        include cross-correlations

    """
    def __init__(self, num_channels, autocorr, rng=None):
        super().__init__(num_channels, rng)
        assert autocorr.shape[0] == num_channels
        
        assert cr.is_autocorr_func(autocorr)
        num_coeffs = []
        denom_coeffs = []
        self.noise_power = np.zeros(self.num_channels)
        for ch_idx in range(self.num_channels):
            ar_coeffs, self.noise_power[ch_idx] = ar_coeffs_from_autocorr(autocorr[ch_idx:ch_idx+1,:])
            nc, dc = ar_coeffs_to_transfer_function(ar_coeffs)
            num_coeffs.append(nc)
            denom_coeffs.append(dc)

        self.filt = fc.IIRFilter(num_coeffs, denom_coeffs)
        self.prepare()

    def prepare(self):
        for ch_idx in range(self.num_channels):
            self.filt.process(self.rng.normal(loc=0, scale=np.sqrt(self.noise_power[ch_idx]), size=(self.num_channels, self.filt.order[ch_idx])))
    
    def get_samples(self, num_samples):
        return self.filt.process(self.rng.normal(loc=0, scale=np.sqrt(self.noise_power)[:,None], size=(self.num_channels, num_samples)))












class MLS(Source):
    """Generates a maximum length sequence"""
    def __init__(self, num_channels, order, polynomial, state=None):
        super().__init__(num_channels, None)
        if num_channels > 1:
            raise NotImplementedError
        self.order = order
        self.poly = polynomial
        self.state = None

        self.metadata["order"] = self.order
        self.metadata["polynomial"] = self.poly
        self.metadata["start state"] = self.state

    def get_samples(self, num_samples):
        seq, self.state = spsig.max_len_seq(
            self.order, state=self.state, length=num_samples, taps=self.poly
        )
        return seq * 2 - 1


class GoldSequenceSource(Source):
    preferredSequences = {
        5: [[2], [1, 2, 3]],
        6: [[5], [1, 4, 5]],
        7: [[4], [4, 5, 6]],
        8: [[1, 2, 3, 6, 7], [1, 2, 7]],
        9: [[5], [3, 5, 6]],
        10: [[2, 5, 9], [3, 4, 6, 8, 9]],
        11: [[9], [3, 6, 9]],
    }

    def __init__(self, num_channels, power, order):
        super().__init__(num_channels)
        assert num_channels <= 2 ** order - 1
        self.num_channels = num_channels
        self.seqLength = 2 ** order - 1
        self.idx = 0
        self.set_power(power)
        
        if isinstance(self.power, (np.ndarray)):
            self.metadata["power"] = self.power.tolist()
        else:
            self.metadata["power"] = self.power
        self.metadata["order"] = order
        self.metadata["sequence length"] = self.seqLength

        if 5 <= order <= 11:
            mls1 = MLS(1, order, self.preferredSequences[order][0])
            mls2 = MLS(1, order, self.preferredSequences[order][1])
            seq1 = mls1.get_samples(2 ** order - 1)
            seq2 = mls2.get_samples(2 ** order - 1)
        elif order > 11:
            assert order % 2 == 1
            mls1 = MLS(1, order, None)
            seq1 = mls1.get_samples(2**order - 1)
            k = util.getSmallestCoprime(order)
            decimationFactor = int(2**k + 1)
            decimatedIndices = (np.arange(self.seqLength) * decimationFactor) % self.seqLength
            seq2 = np.copy(seq1[decimatedIndices])

        self.sequences = np.zeros((num_channels, self.seqLength))
        for i in range(num_channels):
            self.sequences[i, :] = np.negative(
                seq1 * np.roll(seq2, i)
            )  # XOR with integer shifts

    def get_samples(self, num_samples):
        outSignal = np.zeros((self.num_channels, num_samples))
        blockLengths = util.calc_block_sizes(num_samples, self.idx, self.seqLength)
        outIdx = 0
        for blockLen in blockLengths:
            outSignal[:, outIdx : outIdx + blockLen] = self.sequences[
                :, self.idx : self.idx + blockLen
            ]
            outIdx += blockLen
            self.idx = (self.idx + blockLen) % self.seqLength
        return outSignal * self.amplitude

    def set_power(self, newPower):
        if isinstance(newPower, (int, float)) or newPower.ndim == 0:
            self.power = np.ones((1,1)) * newPower
        elif newPower.ndim == 1:
            self.power = newPower[:, None]
        elif newPower.ndim == 2:
            assert newPower.shape[-1] == 1
            self.power = newPower
        else:
            raise ValueError
        self.amplitude = np.sqrt(self.power)




class LinearChirpSource(Source):
    def __init__(self, num_channels, amplitude, freqRange, samplesToSweep, samplerate, rng=None):
        raise NotImplementedError #change amplitude to power

        super.__init__(num_channels, rng)
        assert len(freqRange) == 2
        self.amp = amplitude
        self.samplerate = samplerate
        self.freqRange = freqRange

        self.samplesToSweep = samplesToSweep
        self.deltaFreq = (freqRange[1] - freqRange[0]) / samplesToSweep
        self.freq = freqRange[0]
        self.freqCounter = 0
        self.phase = self.rng.uniform(low=0, high=2 * np.pi)

    def nextPhase(self):
        self.phase += (self.freq / self.samplerate) + (
            self.deltaFreq / (2 * self.samplerate)
        )
        self.phase = self.phase % 1
        self.freq = self.freq + self.deltaFreq

    def get_samples(self, num_samples):
        noise = np.zeros((1, num_samples))
        samplesLeft = num_samples
        n = 0
        for blocks in range(1 + num_samples // self.samplesToSweep):
            block_size = min(samplesLeft, self.samplesToSweep - self.freqCounter)
            for _ in range(block_size):
                noise[0, n] = np.cos(2 * np.pi * self.phase)
                self.nextPhase()

                n += 1
                self.freqCounter += 1

                if self.samplesToSweep <= self.freqCounter:
                    self.deltaFreq = -self.deltaFreq
                    # if self.deltaFreq > 0:
                    #    self.freq = self.freqRange[0]
                    # else:
                    #    self.freq = self.freqRange[1]
                    self.freqCounter = 0
            samplesLeft -= block_size
        noise *= self.amp
        return noise


import numpy as np
import pytest

from aspsim.simulator import SimulatorSetup
import aspsim.diagnostics.diagnostics as dia
import aspsim.signal.sources as sources
import aspcore.filter as fc

@pytest.fixture(scope="session")
def fig_folder(tmp_path_factory):
    return tmp_path_factory.mktemp("figs")


def _setup_ism(fig_folder, samplerate):
    setup = SimulatorSetup(fig_folder)
    setup.sim_info.samplerate = samplerate
    setup.sim_info.tot_samples =  samplerate
    setup.sim_info.sim_chunk_size = 2*samplerate
    setup.sim_info.sim_buffer = samplerate
    setup.sim_info.export_frequency =  samplerate
    setup.sim_info.save_source_contributions = False
    setup.sim_info.randomized_ism = False
    setup.sim_info.start_sources_before_0 = False

    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [7, 5, 5]
    setup.sim_info.room_center = [0, 0, 0]
    setup.sim_info.rt60 = 0.25
    setup.sim_info.max_room_ir_length = samplerate // 2
    return setup

def test_generated_rirs_for_different_positions_are_not_identical(fig_folder):
    sr = 1000
    rng = np.random.default_rng()

    pos_mic = rng.uniform(-1, 1, size=(2,3)) 
    pos_src = np.array([[0,0,0]])

    sim_setup = _setup_ism(fig_folder, sr)
    sim_setup.add_mics("mic", pos_mic)
    sim_setup.add_controllable_source("src", pos_src)
    sim = sim_setup.create_simulator()
    rir1 = sim.arrays.paths["src"]["mic"][0,0,:]
    rir2 = sim.arrays.paths["src"]["mic"][0,1,:]

    assert not np.allclose(rir1, rir2)

def test_generated_rirs_for_same_position_is_consistent_with_more_than_one_simulator(fig_folder):
    sr = 1000
    rng = np.random.default_rng()

    pos_mic = rng.uniform(-1, 1, size=(1,3))
    pos_src = rng.uniform(-1, 1, size=(1,3))

    sim_setup = _setup_ism(fig_folder, sr)
    sim_setup.add_mics("mic", pos_mic)
    sim_setup.add_controllable_source("src", pos_src)
    sim = sim_setup.create_simulator()
    rir1 = sim.arrays.paths["src"]["mic"]

    sim_setup = _setup_ism(fig_folder, sr)
    sim_setup.add_mics("mic", pos_mic)
    sim_setup.add_controllable_source("src", pos_src)
    sim = sim_setup.create_simulator()
    rir2 = sim.arrays.paths["src"]["mic"]
    assert np.allclose(rir1, rir2)

def test_generated_rir_satisfies_reciprocity(fig_folder):
    sr = 1000
    rng = np.random.default_rng()

    pos1 = np.array([[1,0,0]]) #rng.uniform(-1, 1, size=(1,3))
    pos2 = np.array([[0,1,0]]) #¤rng.uniform(-1, 1, size=(1,3))

    sim_setup = _setup_ism(fig_folder, sr)
    sim_setup.add_mics("mic", pos1)
    #sim_setup.add_free_source("src", pos2, sources.WhiteNoiseSource(1, 1, rng))
    sim_setup.add_controllable_source("src", pos2)
    sim = sim_setup.create_simulator()
    rir1 = sim.arrays.paths["src"]["mic"]

    sim_setup = _setup_ism(fig_folder, sr)
    sim_setup.add_mics("mic", pos2)
    sim_setup.add_controllable_source("src", pos1)
    sim = sim_setup.create_simulator()
    rir2 = sim.arrays.paths["src"]["mic"]

    assert np.allclose(rir1, rir2)






def test_simulation_equals_direct_convolution(fig_folder):
    rng = np.random.default_rng()
    sr = 1000

    setup = SimulatorSetup(fig_folder)
    setup.sim_info.samplerate = sr

    setup.sim_info.tot_samples = sr
    setup.sim_info.export_frequency = setup.sim_info.tot_samples
    setup.sim_info.reverb = "ism"
    setup.sim_info.room_size = [7, 4, 2]
    setup.sim_info.room_center = [2, 0.1, 0.1]
    setup.sim_info.rt60 =  0.25
    setup.sim_info.max_room_ir_length = sr // 2
    setup.sim_info.array_update_freq = 1
    setup.sim_info.randomized_ism = False
    setup.sim_info.auto_save_load = False
    setup.sim_info.sim_buffer = sr // 2
    setup.sim_info.extra_delay = 60
    setup.sim_info.plot_output = "pdf"

    src_sig = rng.normal(size=(1, setup.sim_info.tot_samples))

    setup.add_mics("mic", np.zeros((1,3)))
    setup.add_free_source("src", np.array([[1,0,0]]), sources.Sequence(src_sig))

    sim = setup.create_simulator()
    sim.diag.add_diagnostic("mic", dia.RecordSignal("mic", sim.sim_info, 1, export_func="npz"))
    sim.run_simulation()

    sig_sim = np.load(sim.folder_path.joinpath(f"mic_{sim.sim_info.tot_samples}.npz"))["mic"]

    filt = fc.create_filter(ir=sim.arrays.paths["src"]["mic"])
    direct_mic_sig = filt.process(src_sig)

    assert np.allclose(sig_sim, direct_mic_sig)


import numpy as np


def calc_block_sizes(num_samples, start_idx, block_size):
    left_in_block = block_size - start_idx
    sample_counter = 0
    block_sizes = []
    while sample_counter < num_samples:
        block_len = np.min((num_samples - sample_counter, left_in_block))
        block_sizes.append(block_len)
        sample_counter += block_len
        left_in_block -= block_len
        if left_in_block == 0:
            left_in_block = block_size
    return block_sizes


import numpy as np





class PhaseCounter:
    """An index counter to keep track of non-overlapping continous phases. 

    The number is how many samples that each phase should be. The first phase will start at sample 0.

    Example
    -------
    A processor needs the first 2000 samples for an initialization, then must wait 5000 samples before beginning the real processing step.
    The class can be used by defining the phases as
    >>> phase_def = {'init' : 2000, 'wait' : 5000, 'process' : np.inf}

    and then checking if either of the following is true
    >>> if phase_counter.phase == 'init'
    >>> if phase_counter.current_phase_is('init')

    Notes
    -----
    np.inf represents an infinite length
    This should naturally only be used for the last phase
    If all phases has finished, the phase will be None. 

    first_sample will be True on the first sample of each phase,
    allowing running one-time functions in each phase

    Extended implementation to block_size != 1 can be done later

    """
    def __init__(self, phase_lengths, verbose=False):
        assert isinstance(phase_lengths, dict)
        self.phase_lengths = phase_lengths
        self.verbose = verbose
        self.phase = None
        self.first_sample = True
        

        #phase_lengths = {name : length for name, length in self.phase_lengths.items() if length != 0}
        #phase_lengths = {name : length for name, length in self.phase_lengths.items()}
        
        #phase_idxs = [i for i in self.phase_lengths.values() if i != 0]
        self.phase_lengths = {name : i if i >= 0 else np.inf for name, i in self.phase_lengths.items()}
        #assert all([i != 0 for i in p_len])
        self.start_idxs = np.cumsum(list(self.phase_lengths.values())).tolist()
        self.start_idxs = [i if np.isinf(i) else int(i) for i in self.start_idxs]
        self.start_idxs.insert(0,0)

        self.phase_names = list(self.phase_lengths.keys())
        if self.start_idxs[-1] < np.inf:
            self.phase_names.append(None)
        else:
            self.start_idxs.pop()

        self.start_idxs = {phase_name:start_idx for phase_name, start_idx in zip(self.phase_names, self.start_idxs)}

        self._phase_names = [phase_name for phase_name, phase_len in self.phase_lengths.items() if phase_len > 0]
        self._start_idxs = [start_idx for start_idx, phase_len in zip(self.start_idxs.values(), self.phase_lengths.values()) if phase_len > 0]

        self.idx = 0
        self.next_phase()

    def next_phase(self):
        if self.verbose:
            print(f"Changed phase from {self.phase}")
            
        self.phase = self._phase_names.pop(0)
        self._start_idxs.pop(0)
        if len(self._start_idxs) == 0:
            self._start_idxs.append(np.inf)
        self.first_sample = True
        
        if self.verbose:
            print(f"to {self.phase}")

    def progress(self):
        self.idx += 1
        if self.idx >= self._start_idxs[0]:
            self.next_phase()
        else:
            self.first_sample = False

    def current_phase_is(self, phase_name):
        return self.phase == phase_name



class EventCounter:
    """
    An index counter to keep track of events that should 
    only happen every x samples

    event_def is a dictionary with all event
    each entry is event_name : (frequency, offset)

    Example:
    event_counter = EventCounter({'event_1' : (256,0), 'event_2' : (1,0), 'event_3' : (1024,256)})
    event_2 will happen every sample, event_1 every 256 samples
    First at sample 256 all three events will happen simultaneouly. 

    To be used as:
    if event_name in event_counter.event:
        do_thing()

    """
    def __init__(self, event_def):
        self.event_def = event_def
        self.event = []

        self.freq = {name : freq for name, (freq, offset) in event_def.items()}
        self.offset = {name : offset for name, (freq, offset) in event_def.items()}

        self.idx = 0

    def add_event(self, name, freq, offset):
        self.event_def[name] = (freq, offset)

    def _check_events(self):
        self.event = []
        for name, (freq, offset) in self.event_def.items():
            if (self.idx - offset) % freq == 0:
                self.event.append(name)

    def progress(self):
        self.idx += 1 
        self._check_events()


import numpy as np
import scipy.signal as spsig



# =================== SCALING ===================
def linear(signal):
    return signal

def db_power(signal):
    return 10*np.log10(signal)

def db_amplitude(signal):
    return 20*np.log10(signal)

def natural_log(signal):
    return np.log(signal)

# def preset_scaling(scaling):
#     if scaling == "linear":
#         return linear
#     elif scaling == "db_power":
#         return db_power
#     elif scaling == "db_amplitude":
#         return db_amplitude
#     elif scaling == "natural_log":
#         return natural_log
#     else:
#         raise ValueError("Invalid scaling string")

# def scale_data(signal, scaling):
#     if callable(scaling):
#         return scaling(signal)
#     elif isinstance(scaling, str):
#         scaling_func = preset_scaling(scaling)
#         return scaling_func(signal)
#     else:
#         raise ValueError("Invalid scaling type")


def clip(low_lim, high_lim):
    def clip_internal(signal):
        return np.clip(signal, low_lim, high_lim)
    return clip_internal


def smooth(smooth_len):
    ir = np.ones((1, smooth_len)) / smooth_len
    def smooth_internal(signal):
        return spsig.oaconvolve(signal, ir, mode="full", axes=-1)[:,:signal.shape[-1]]
    return smooth_internal




import numpy as np
import hypothesis as hyp
import hypothesis.strategies as st
import pytest

from aspsim.simulator import SimulatorSetup
import aspsim.processor as bse
import aspsim.diagnostics.diagnostics as dia
import aspsim.signal.sources as sources
import aspcore.filter as fc

import aspsim.configutil as cu

def _default_sim_info():

    sim_info = cu.load_default_config()
    sim_info.tot_samples = 20
    sim_info.sim_buffer = 20
    sim_info.export_frequency = 20
    sim_info.sim_chunk_size = 20
    sim_info.max_room_ir_length = 8

    sim_info.start_sources_before_0 = False
    return sim_info


@pytest.fixture(scope="session")
def fig_folder(tmp_path_factory):
    return tmp_path_factory.mktemp("figs")

def _setup_with_debug_processor(fig_folder):
    setup = SimulatorSetup(fig_folder)
    setup.sim_info.tot_samples = 20
    setup.sim_info.sim_buffer = 20
    setup.sim_info.export_frequency = 20
    setup.sim_info.sim_chunk_size = 20

    setup.add_free_source("src", np.array([[1,0,0]]), sources.WhiteNoiseSource(1,1))
    setup.add_controllable_source("loudspeaker", np.array([[1,0,0]]))
    setup.add_mics("mic", np.zeros((1,3)))

    setup.arrays.path_type["loudspeaker"]["mic"] = "none"
    setup.arrays.path_type["src"]["mic"] = "direct"
    return setup

@pytest.fixture(scope="session")
def _sim_setup(tmp_path_factory):
    setup = SimulatorSetup(tmp_path_factory.mktemp("figs"), None)
    return setup


@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5))
def test_minimum_of_tot_samples_are_processed(fig_folder, bs):
    sim_setup = _setup_with_debug_processor(fig_folder)
    sim = sim_setup.create_simulator()
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs))
    sim.run_simulation()
    assert sim.processors[0].processed_samples >= sim.sim_info.tot_samples

@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5))
def test_consecutive_simulators_give_same_values(fig_folder, bs):
    # change this to a free source instead without processors
    sim_setup = _setup_with_debug_processor(fig_folder)
    sim = sim_setup.create_simulator()
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs))
    sim.run_simulation()

    sig1 = sim.sig["mic"]

    sim = sim_setup.create_simulator()
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs))
    sim.run_simulation()

    sig2 = sim.sig["mic"]
    assert np.allclose(sig1, sig2)

@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5))
def test_minimum_value_for_sim_buffer(fig_folder, bs):
    assert False # not implemented yet
    _sim_setup = _setup_with_debug_processor(fig_folder)
    sim = _sim_setup.create_simulator()
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs))
    sim.run_simulation()
    assert sim.processors[0].processed_samples >= sim.sim_info.tot_samples

@hyp.settings(deadline=None)
@hyp.given(bs = st.integers(min_value=1, max_value=5))
def test_correct_processing_delay(fig_folder, bs):
    sim_setup = _setup_with_debug_processor(fig_folder)
    sim = sim_setup.create_simulator()
    sim.add_processor(bse.DebugProcessor(sim.sim_info, sim.arrays, bs))
    sim.run_simulation()

    proc = sim.processors[0]
    assert np.allclose(proc.mic[:,:-bs], proc.ls[:,bs:])

def test_simulation_equals_direct_convolution_random_rirs(fig_folder):
    rng = np.random.default_rng()
    setup = SimulatorSetup(fig_folder)
    setup.sim_info = _default_sim_info()

    src_sig = rng.normal(size=(1, setup.sim_info.tot_samples))

    setup.add_mics("mic", np.zeros((1,3)))
    setup.add_free_source("src", np.array([[1,0,0]]), sources.Sequence(src_sig))
    setup.arrays.path_type["src"]["mic"] = "random"

    sim = setup.create_simulator()
    sim.diag.add_diagnostic("mic", dia.RecordSignal("mic", sim.sim_info, 1, export_func="npz"))
    sim.run_simulation()

    sig_sim = np.load(sim.folder_path.joinpath(f"mic_{sim.sim_info.tot_samples}.npz"))["mic"]

    filt = fc.create_filter(ir=sim.arrays.paths["src"]["mic"])
    direct_mic_sig = filt.process(src_sig)

    assert np.allclose(sig_sim, direct_mic_sig)

@hyp.settings(deadline=None)
@hyp.given(num_src = st.integers(min_value=1, max_value=3))
def test_simulation_equals_direct_convolution_random_rirs_multiple_sources(fig_folder, num_src):
    rng = np.random.default_rng()
    setup = SimulatorSetup(fig_folder)
    setup.sim_info = _default_sim_info()

    src_sig = [rng.normal(size=(1, setup.sim_info.tot_samples)) for s in range(num_src)]

    setup.add_mics("mic", np.zeros((1,3)))
    for s in range(num_src):    
        setup.add_free_source(f"src_{s}", np.zeros((1,3)), sources.Sequence(src_sig[s]))
        setup.arrays.path_type[f"src_{s}"]["mic"] = "random"

    sim = setup.create_simulator()
    sim.diag.add_diagnostic("mic", dia.RecordSignal("mic", sim.sim_info, 1, export_func="npz"))
    sim.run_simulation()

    sig_sim = np.load(sim.folder_path.joinpath(f"mic_{sim.sim_info.tot_samples}.npz"))["mic"]

    filt = [fc.create_filter(ir=sim.arrays.paths[f"src_{s}"]["mic"]) for s in range(num_src)]
    direct_mic_sig = np.sum(np.concatenate([filt[s].process(src_sig[s]) for s in range(num_src)], axis=0), axis=0)

    assert np.allclose(sig_sim, direct_mic_sig)










